from abc import ABC
import numpy as np
import scipy as sp
from math import pi, sin, cos
import matplotlib.pyplot as plt
from spatialmath import base
import roboticstoolbox as rtb


"""
Sensor Sensor superclass

An abstract superclass to represent robot navigation sensorself.

Methods::
  plot        plot a line from robot to map feature
  display     print the parameters in human readable form
  char        convert to string

Properties::
robot   The Vehicle object on which the sensor is mounted
map     The PointMap object representing the landmarks around the robot

Reference::

  Robotics, Vision & Control,
  Peter Corke,
  Springer 2011

See also RangeBearingSensor, EKF, Vehicle, Landmarkself.

"""

class Sensor(ABC):
    # TODO, pose option, wrt vehicle

        # robot
        # map
        
        # verbose
        
        # ls
        # animate     # animate sensor measurements
        # interval    # measurement return subsample factor
        # fail
        # delay
        

    def __init__(self, robot, map, every=1, fail=[], animate=False, delay=0.1, verbose=False):
        """Sensor.Sensor Sensor object constructor
        %
        # S = Sensor(VEHICLE, MAP, OPTIONS) is a sensor mounted on a vehicle
        # described by the Vehicle subclass object VEHICLE and observing landmarks
        # in a map described by the LandmarkMap class object self.
        %
        # Options::
        # 'animate'    animate the action of the laser scanner
        # 'ls',LS      laser scan lines drawn with style ls (default 'r-')
        # 'skip', I    return a valid reading on every I'th call
        # 'fail',T     sensor simulates failure between timesteps T=[TMIN,TMAX]
        %
        # Notes::
        # - Animation shows a ray from the vehicle position to the selected
        #   landmark.
        """
        self._robot = robot
        self._map = map
        self._every = every
        self._fail =  fail
        
        self._count = 1
        self._verbose = verbose

        self.delay = 0.1

        self._animate = animate

        self._rand = np.random.default_rng()


    def __str__(self):
        """
        Convert sensor parameters to a string
        %
        # s = self.char() is a string showing sensor parameters in
        # a compact human readable format.
        """
        s = f"{self.__class__.__name__} sensor class\n"
        s += "  " + str(self.map)
        return s

    @property
    def robot(self):
        return self._robot

    @property
    def map(self):
        return self._map

    @property
    def sensor(self):
        return self._sensor

    @property
    def verbose(self):
        return self._verbose
    
    def plot(s, jf):
        """Sensor.plot Plot sensor reading
        %
        # self.plot(J) draws a line from the robot to the J'th map feature.
        %
        # Notes::
        # - The line is drawn using the linestyle given by the property ls
        # - There is a delay given by the property delay
        """
        # if isempty(self.ls)
        #     return
        # end
        
        # h = findobj(gca, 'tag', 'sensor')
        # if isempty(h)
        #     # no sensor line, create one
        #     h = plot(0, 0, self.ls, 'tag', 'sensor')
        # end
        
        # # there is a sensor line animate it
        
        # if jf == 0
        #     set(h, 'Visible', 'off')
        # else
        #     xi = self.self.map(:,jf)
        #     set(h, 'Visible', 'on', 'XData', [self.robot.x[0], xi[0]], 'YData', [self.robot.x[1], xi[1]])
        # end
        # pause(self.delay)

        # drawnow
        pass



"""RangeBearingSensor Range and bearing sensor class

A concrete subclass of the Sensor class that implements a range and bearing
angle sensor that provides robot-centric measurements of landmark points in 
the world. To enable this it holds a references to a map of the world (LandmarkMap object)
and a robot (Vehicle subclass object) that moves in SE[1].

The sensor observes landmarks within its angular field of view between
the minimum and maximum range.

Methods::

reading   range/bearing observation of random landmark
h         range/bearing observation of specific landmark
Hx        Jacobian matrix with respect to vehicle pose dh/dx 
Hp        Jacobian matrix with respect to landmark position dh/dp 
Hw        Jacobian matrix with respect to noise dh/dw
-
g         feature position given vehicle pose and observation
Gx        Jacobian matrix with respect to vehicle pose dg/dx 
Gz        Jacobian matrix with respect to observation dg/dz

Properties (read/write)::
W            measurement covariance matrix (2x2)
interval     valid measurements returned every interval'th call to reading()
landmarklog  time history of observed landmarks

Reference::

  Robotics, Vision & Control, Chap 6,
  Peter Corke,
  Springer 2011

See also Sensor, Vehicle, LandmarkMap, EKF."""

# ======================================================================== #

class RangeBearingSensor(Sensor):

    # properties
    #     W           # measurment covariance
    #     r_range     # range limits
    #     theta_range # angle limits

    #     randstream  # random stream just for Sensors
        
    #     landmarklog  # time history of observed landmarks        
    # end

    # properties (SetAccess = private)
    #     count       # number of reading()s
    # end


    def __init__(self, robot, map, 
            line_style=None,
            poly_style=None,
            covar=None, 
            range=None,
            angle=None,
            **kwargs):
        """RangeBearingSensor.RangeBearingSensor Range and bearing sensor constructor
        %
        # S = RangeBearingSensor(VEHICLE, MAP, OPTIONS) is an object
        # representing a range and bearing angle sensor mounted on the Vehicle
        # subclass object VEHICLE and observing an environment of known landmarks
        # represented by the LandmarkMap object self.  The sensor covariance is W
        # (2x2) representing range and bearing covariance.
        %
        # The sensor has specified angular field of view and minimum and maximum
        # range.
        %
        # Options::
        # 'covar',W               covariance matrix (2x2)
        # 'range',xmax            maximum range of sensor
        # 'range',[xmin xmax]     minimum and maximum range of sensor
        # 'angle',TH              angular field of view, from -TH to +TH
        # 'angle',[THMIN THMAX]   detection for angles betwen THMIN
        #                         and THMAX
        # 'skip',K                return a valid reading on every K'th call
        # 'fail',[TMIN TMAX]      sensor simulates failure between 
        #                         timesteps TMIN and TMAX
        # 'animate'               animate sensor readings
        %
        # See also options for Sensor constructor.
        %
        # See also RangeBearingSensor.reading, Sensor.Sensor, Vehicle, LandmarkMap, EKF.
        """


        # call the superclass constructor
        super().__init__(robot, map, **kwargs)

        self._line_style = line_style
        self._poly_style = poly_style

        if covar is None:
            self._covar = np.zeros((2,2))
        else:
            self._covar = covar
        self._W = self._covar

        if range is None:
            self._r_range = None
        elif len(range) == 1:
            self._r_range = [0, range]
        elif len(range) == 2:
            self._r_range = range

        if angle is None:
            self._theta_range = None
        elif len(angle) == 1:
            self._theta_range = [-angle, angle]
        elif len(angle) == 2:
            self._theta_range = angle

        self._landmarklog = []


    def init(self):
        self._landmarklog = []
        self._count = 1
    
    def selectFeature(self):
        return self.randstream.randi(self._sensor.nlandmarks)


    def reading(self):
        """
        Choose landmark and return observation
        %
        # [Z,K] = self.reading() is an observation of a random visible landmark where
        # Z=[R,THETA] is the range and bearing with additive Gaussian noise of
        # covariance W (property W). K is the index of the map feature that was
        # observed.
        %
        # The landmark is chosen randomly from the set of all visible landmarks,
        # those within the angular field of view and range limitself.  If no valid
        # measurement, ie. no features within range, interval subsampling enabled
        # or simulated failure the return is Z=[] and K=0.
        %
        # Notes::
        # - Noise with covariance W (property W) is added to each row of Z.
        # - If 'animate' option set then show a line from the vehicle to the
        #   landmark
        # - If 'animate' option set and the angular and distance limits are set
        #   then display that region as a shaded polygon.
        # - Implements sensor failure and subsampling if specified to constructor.
        %
        # See also RangeBearingSensor.h.
        """
        
        # TODO probably should return K=0 to indicated invalid
        
        # model a sensor that emits readings every interval samples
        self._count += 1

        # check conditions for NOT returning a value
        z = []
        jf = 0
        # sample interval
        if self._count % self._every != 0:
            return (None, None)

        # simulated failure, fail is a list of 2-tuples giving (start,end) times
        # for a sensor failure
        if self._fail is not None:
            if any([start <= self._count < end for start, end in self._fail]):
                return (None, None)
        
        # create a polygon to indicate the active sensing area based on range+angle limits
        # if self.animate && ~isempty(self.theta_range) && ~isempty(self.r_range)
        #     h = findobj(gca, 'tag', 'sensor-area')
        #     if isempty(h)
                
        #         th=linspace(self.theta_range[0], self.theta_range[1], 20)
        #         x = self.r_range[1] * cos(th)
        #         y = self.r_range[1] * sin(th)
        #         if self.r_range[0] > 0
        #             th = flip(th)
        #             x = [x self.r_range[0] * cos(th)]
        #             y = [y self.r_range[0] * sin(th)]
        #         else
        #             x = [x 0]
        #             y = [y 0]
        #         end
        #         # no sensor zone, create one
        #         plot_poly([x; y], 'fillcolor', 'r', 'alpha', 0.1, 'edgecolor', 'none', 'animate', 'tag', 'sensor-area')
        #     else
        #         hg = get(h, 'Parent')
        #         plot_poly(h, self.robot.x)


        # if range and bearing angle limits are in place look for
        # any landmarks that match criteria

        # get range/bearing to all landmarks, one per row
        z = self.h(self.robot.x)
        zk = [(z, k) for k, z in enumerate(z)]
        # a list of tuples, each tuple is ((range, bearing), k)
        
        if self._r_range is not None:
            zk = filter(lambda zk: self._r_range[0] <= zk[0][0] <= self._r_range[1], zk)

        if self._theta_range is not None:
            # find all within angular range as well
            zk = filter(lambda zk: self._theta_range[0] <= zk[0][1] <= self._theta_range[1], zk)
            
        if len(zk) > 1:
            # more than 1 in range, pick a random one
            i = self._rand.integers(len(zk))
            z = zk[i][0]
            jf = zk[i][1]
            if self.verbose:
                print(f"Sensor:: feature {jf}: ({z[0]}, {z[1]})")
        else:
            if self.verbose:
                print('Sensor:: no features\n')
            return (None, None)

        # compute the range and bearing from robot to feature
        # z = self.h(self.robot.x, jf)  
        
        if self._animate:
            self.plot(jf)
        
        # add the reading to the landmark log
        self._landmarklog.append(jf)
    
        return z


    def h(self, xv, jf=None):
        """
        Landmark range and bearing
        %
        # Z = self.h(X, K) is a sensor observation (1x2), range and bearing, from vehicle at 
        # pose X (1x3) to the K'th landmark.
        %
        # Z = self.h(X, P) as above but compute range and bearing to a landmark at coordinate P.
        %
        # Z = self.h(X) as above but computes range and bearing to all
        # map featureself.  Z has one row per landmark.
        %
        # Notes::
        # - Noise with covariance W (propertyW) is added to each row of Z.
        # - Supports vectorized operation where XV (Nx3) and Z (Nx2).
        # - The landmark is assumed visible, field of view and range liits are not
        #   applied.
        %
        # See also RangeBearingSensor.reading, RangeBearingSensor.Hx, RangeBearingSensor.Hw, RangeBearingSensor.Hp.
        """
        # get the landmarks, one per row
        if jf is None:
            # self.h(XV)   all landmarks
            dx = self.map.x - xv[0]
            dy = self.map.y - xv[1]
        elif base.isinteger(jf):
            # self.h(XV, JF)
            xlm = self.map.landmark(jf)
            dx = xlm[0] - xv[0]
            dy = xlm[1] - xv[1]
        else:
            # self.h(XV, XF)
            xlm = base.getvector(jf, 2)
            dx = xlm[0] - xv[0]
            dy = xlm[1] - xv[1]

        # compute range and bearing (Vectorized code)

        z = np.c_[
                np.sqrt(dx**2 + dy**2), 
                base.angdiff(np.arctan2(dy, dx), xv[2]) 
            ]  # range & bearing as columns

        # add noise with covariance W
        z += self._rand.normal(size=z.shape) @ sp.linalg.sqrtm(self._W)

        if z.shape[0] == 1:
            return z[0]
        else:
            return z

    def Hx(self, xv, jf):
        """
        Jacobian dh/dx
        %
        # J = self.Hx(X, K) returns the Jacobian dh/dx (2x3) at the vehicle
        # state X (3x1) for map landmark K.
        %
        # J = self.Hx(X, P) as above but for a landmark at coordinate P.
        %
        # See also RangeBearingSensor.h.
        """
        if base.isinteger(jf):
            # landmark index provided
            xf = self.sensor.landmark(jf)
        else:
            # assume it is a coordinate
            xf = base.getvector(jf, 2)

        Delta = xf - xv[0:2]
        r = base.norm(Delta)
        return np.array([
            [-Delta[0] / r,    -Delta[1] / r,      0],
            [ Delta[1] / r**2, -Delta[0] / r**2,  -1],
                ])

    def Hp(self, xv, jf):
        """
        Jacobian dh/dp
        %
        # J = self.Hp(X, K) is the Jacobian dh/dp (2x2) at the vehicle
        # state X (3x1) for map landmark K.
        %
        # J = self.Hp(X, P) as above but for a landmark at coordinate P (1x2).
        %
        # See also RangeBearingSensor.h.
        """
        if base.isinteger(jf):
            xf = self.map.landmark(jf)
        else:
            xf = jf

        Delta = xf - xv[0:2]
        r = base.norm(Delta)
        return np.array([
            [ Delta([0]) / r,      Delta([1]) / r],
            [-Delta([1]) / r**2,    Delta([0]) / r**2],
            ])

    def Hw(self, xv, jf):
        """
        Jacobian dh/dw
        %
        # J = self.Hw(X, K) is the Jacobian dh/dw (2x2) at the vehicle
        # state X (3x1) for map landmark K.
        %
        # See also RangeBearingSensor.h.
        """
        return np.eye(2)

    def g(self, xv, z):
        """
        Compute landmark location
        %
        # P = self.g(X, Z) is the world coordinate (2x1) of a feature given
        # the observation Z (1x2) from a vehicle state with X (3x1).
        %
        # See also RangeBearingSensor.Gx, RangeBearingSensor.Gz.
        """
        range = z[0]
        bearing = z[1] + xv[2]  # bearing angle in vehicle frame

        return np.r_[
            xv[0] + range * cos(bearing),
            xv[1] + range * sin(bearing)
                ]

    def Gx(self, xv, z):
        """RangeBearingSensor.Gxv Jacobian dg/dx
        %
        # J = self.Gx(X, Z) is the Jacobian dg/dx (2x3) at the vehicle state X (3x1) for
        # sensor observation Z (2x1).
        %
        # See also RangeBearingSensor.g.
        """
        theta = xv[2]
        r = z[0]
        bearing = z[1]

        return np.array([
            [1,   0,   -r*sin(theta + bearing)],
            [0,   1,    r*cos(theta + bearing)],
                ])

    def Gz(self, xv, z):
        """
        Jacobian dg/dz
        %
        # J = self.Gz(X, Z) is the Jacobian dg/dz (2x2) at the vehicle state X (3x1) for
        # sensor observation Z (2x1).
        %
        # See also RangeBearingSensor.g.
        """
        theta = xv[2]
        r = z[0]
        bearing = z[1]
        return np.array([
            [cos(theta + bearing),   -r*sin(theta + bearing)],
            [sin(theta + bearing),    r*cos(theta + bearing)],
            ])

    def __str__(self):
        s = super().__str__()
        s += f"\n  W = {self._W.tolist()}\n"

        s += f"  sampled every {self._every} samples\n"
        if self._r_range is not None:
            s += f"  range: {self._r_range[0]} to {self._r_range[0]}\n"
        if self._theta_range is not None:
            s += f"  angle: {self._theta_range[0]} to {self._theta_range[0]}\n"
        return s.rstrip()

if __name__ == "__main__":
    
    pass
    


# """
# Defines the object 'arm' in the current workspace

# Also define the vector qz = [pi/4,0,-pi/3]
# """

# from roboticstoolbox.robot.serial_link import *


# class Uprighttl(SerialLink):

#     def __init__(self):
#         L = [Link(a=0.1, alpha=pi / 2, jointtype='R'),
#              Link(a=1.0, jointtype='R'),
#              Link(a=1.0, jointtype='R'),
#              Link(a=0.5, jointtype='R')]

#         self._qz = [0, 0, 0, 0]

#         super(Uprighttl, self).__init__(L, name='Upright Arm')

#     @property
#     def qz(self):
#         return self._qz



#!/usr/bin/env python
"""
@author: Peter Corke
@author: Samuel Drew
"""

from roboticstoolbox import DHRobot, RevoluteDH
from math import pi
import numpy as np


class Ball(DHRobot):
    """
    Class that models a ball manipulator

    :param N: number of links, defaults to 10
    :type N: int, optional
    :param symbolic: [description], defaults to False
    :type symbolic: bool, optional

    The ball robot is an *abstract* robot with an arbitrary number of joints.
    At zero joint angles it is straight along the x-axis, and as the joint
    angles are increased (equally) it wraps up into a 3D ball shape.

    - ``Ball()`` is an object which describes the kinematic characteristics of
      a ball robot using standard DH conventions.

    - ``Ball(N)`` as above, but models a robot with ``N`` joints.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Ball()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero joint angles
        - q1, ball shaped configuration

    :references:

        - "A divide and conquer articulated-body algorithm for parallel
          O(log(n)) calculation of rigid body dynamics, Part 2",
          Int. J. Robotics Research, 18(9), pp 876-892.

    :seealso: :func:`Hyper`, :func:`Ball`

    .. codeauthor:: Peter Corke
    """

    def __init__(self, N=10):

        links = []
        q1 = []

        for i in range(N):
            links.append(RevoluteDH(a=0.1, alpha=pi / 2))

        # and build a serial link manipulator
        super(Ball, self).__init__(links, name='ball')

        # zero angles, ball pose
        self.addconfiguration("qz", np.zeros(N,))
        self.addconfiguration("q1", [_fract(i) for i in range(N)])

def _fract(i):
    # i is "i-1" as per the paper
    theta1 = 1
    theta2 = -2/3

    if i == 0:
        return theta1
    elif i % 3 == 1:
        return theta1
    elif i % 3 == 2:
        return theta2
    elif i % 3 == 0:
        return _fract(i / 3)

if __name__ == '__main__':   # pragma nocover

    ball = Ball()
    print(ball)



import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class UR5(DHRobot):
    """
    Class that models a Universal Robotics UR5 manipulator

    :param symbolic: use symbolic constants
    :type symbolic: bool

    ``UR5()`` is an object which models a Unimation Puma560 robot and
    describes its kinematic and dynamic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.UR5()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration
    - qr, arm horizontal along x-axis

    .. note::
        - SI units are used.

    :References:

        - `Parameters for calculations of kinematics and dynamics <https://www.universal-robots.com/articles/ur/parameters-for-calculations-of-kinematics-and-dynamics>`_
        
    :sealso: :func:`UR4`, :func:`UR10`


    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0

        deg = pi / 180
        inch = 0.0254

        # robot length values (metres)
        a = [0, -0.42500, -0.39225, 0, 0, 0]
        d = [0.089459, 0, 0, 0.10915, 0.09465, 0.0823]

        alpha = [pi/2, zero, zero, pi/2, -pi/2, zero]

        # mass data, no inertia available
        mass = [3.7000, 8.3930, 2.33, 1.2190, 1.2190, 0.1897]
        center_of_mass = [
                [0,     -0.02561,  0.00193],
                [0.2125, 0,        0.11336],
                [0.15,   0,        0.0265],
                [0,     -0.0018,   0.01634],
                [0,     -0.0018,   0.01634],
                [0,      0,       -0.001159]
            ]
        links = []

        for j in range(6):
            link = RevoluteDH(
                d=d[j],
                a=a[j],
                alpha=alpha[j],
                m=mass[j],
                r=center_of_mass[j],
                G=1
            )
            links.append(link)
    
        super().__init__(
            links,
            name="UR5",
            manufacturer="Universal Robotics",
            keywords=('dynamics', 'symbolic'),
            symbolic=symbolic
        )
    
        # zero angles
        self.addconfiguration("qz", np.array([0, 0, 0, 0, 0, 0]))
        # horizontal along the x-axis
        self.addconfiguration("qr", np.r_[180, 0, 0, 0, 90, 0]*deg)

if __name__ == '__main__':    # pragma nocover

    ur5 = UR5(symbolic=False)
    print(ur5)
    print(ur5.dyntable())



#!/usr/bin/env python
"""
@author: Peter Corke
"""


from math import pi
import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Coil(DHRobot):
    """
    Create model of a coil manipulator

    :param N: number of links, defaults to 10
    :type N: int, optional
    :param symbolic: [description], defaults to False
    :type symbolic: bool, optional

    The coil robot is an *abstract* robot with an arbitrary number of joints
    that folds into a helix shape.  At zero joint angles it is straight along
    the x-axis, and as the joint angles are increased (equally) it wraps up into
    a 3D helix shape.

    - ``Coil()`` is an object which describes the kinematic characteristics of
      a coil robot using standard DH conventions.

    - ``Coil(N)`` as above, but models a robot with ``N`` joints.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Coil()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration

    :references:
        - "A divide and conquer articulated-body algorithm for parallel O(log(n))
          calculation of rigid body dynamics, Part 2",
          Int. J. Robotics Research, 18(9), pp 876-892. 

    :seealso: :func:`Hyper`, :func:`Ball`

    .. codeauthor:: Peter Corke
    """  # noqa

    def __init__(self, N=10, symbolic=False):

        if symbolic:
            import spatialmath.base.symbolic as sym
            zero = sym.zero()
            pi = sym.pi()
        else:
            from math import pi
            zero = 0.0
            
        a = 1 / N

        links = []
        for i in range(N):
            links.append(RevoluteDH(a=a, alpha=5 * pi / N))

        super().__init__(
            links,
            name="Hyper" + str(N),
            keywords=('symbolic',),
            symbolic=symbolic
        )

        # zero angles, straight
        self.addconfiguration("qz", np.zeros((N,)))

        # folded, helix with ~5.5 turns
        self.addconfiguration("qf", np.ones((N,)) * 10 * pi / N)

if __name__ == '__main__':    # pragma nocover

    coil = Coil(N=10, symbolic=False)
    print(coil)

    # print(coil.fkine(coil.qz))
    # print(coil.fkine(coil.qf))


from roboticstoolbox.backends.PyPlot import *   # noqa
from roboticstoolbox.backends.Swift import *   # noqa
from roboticstoolbox.backends.ROS import *   # noqa

try:
    from roboticstoolbox.backends.VPython import *   # noqa
except ImportError:    # pragma nocover
    pass



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.ERobot import ERobot
from roboticstoolbox.robot.ELink import ELink


class Planar_Y(ERobot):
    """
    Create model of a branched planar manipulator::

        L0 -- L1 -+- L2a -- L3a -- EEa
                |
                +- L2b -- L3b -- EEb

    ``Planar_Y()`` creates a planar branched manipulator model.


    :references:
        - Kinematic Derivatives using the Elementary Transform
          Sequence, J. Haviland and P. Corke

    """
    def __init__(self):

        deg = np.pi/180
        mm = 1e-3
        tool_offset = 1

        # trunk of the tree
        l0 = ELink(
            ETS.rz(),
            name='link0',
            jindex=0,
            parent=None
        )

        l1 = ELink(
            ETS.tx(1) * ETS.rz(),
            name='link1',
            jindex=1,
            parent=l0
        )

        # branch 1
        l2 = ELink(
            ETS.tx(1) * ETS.rz(),
            name='link2a',
            jindex=2,
            parent=l1
        )

        l3 = ELink(
            ETS.tx(1) * ETS.rz(),
            name='link3a',
            jindex=3,
            parent=l2
        )

        eea = ELink(
            ETS.tz(tool_offset),
            name='eea',
            parent=l3
        )

        # branch 2
        l4 = ELink(
            ETS.tx(1) * ETS.rz(),
            name='link2b',
            jindex=4,
            parent=l1
        )

        l5 = ELink(
            ETS.tx(1) * ETS.rz(),
            name='link3b',
            jindex=5,
            parent=l4
        )

        eeb = ELink(
            ETS.tz(tool_offset),
            name='eeb',
            parent=l5
        )

        elinks = [l0, l1, l2, l3, l4, l5, eea, eeb]

        super().__init__(
            elinks,
            name='Planar-Y'
            )

        self.addconfiguration(
            "qz", [0, 0, 0, 0, 0, 0])
        self.addconfiguration(
            "qy", [0, 0, np.pi/4, 0, -np.pi/4, 0])


if __name__ == '__main__':   # pragma nocover

    robot = Planar_Y()
    print(robot)





"""
Python ReedShepp Planner
@Author: Kristian Gibson
TODO: Comments + Sphynx Docs Structured Text
TODO: Bug-fix, testing

Not ready for use yet.
"""
from numpy import disp
from scipy import integrate
from spatialmath.base.transforms2d import *
from spatialmath.base.vectors import *
from spatialmath.pose2d import SE2
from scipy.ndimage import *
from matplotlib import cm
from roboticstoolbox.mobile.navigation import Navigation

class DStar(Navigation):
    def __init__(self, occ_grid=None, 
                 reset=False, **kwargs):
        self._g = None
        self._b = None
        self._t = None
        self._h = None
        self._valid_plan = None
        self._open_list = None
        self._open_list_maxlen = None
        self._new = 0
        self._open = 1
        self._closed = 2
        self._niter = None
        self._costmap = occ_grid.astype(np.float32)
        self._m = 'euclidean'

        super().__init__(occ_grid=occ_grid, **kwargs)

    @property
    def niter(self):
        return self._niter

    @property
    def costmap(self):
        return self._costmap

    def dstar(self, map, goal=None, metric=None, inflate=None, progress=None):
        self._occ_grid = map
        self._goal = goal

        self.reset()
        if self._goal is not None or self._goal is not np.array([]):
            self.goal_change()

        self.reset()

        return self

    def reset(self):
        self._b = np.zeros(self._costmap.shape)
        self._t = np.zeros(self._costmap.shape)
        self._h = np.inf * np.ones(self._costmap.shape)
        
        self._open_list = np.zeros(2, 0)

        self._open_list_maxlen = -np.inf

        self.occ_grid2costmap(self._occ_grid_nav)

        self._valid_plan = False

    def __str__(self):
        s = super().__str__()
        s += f"\n  Distance metric: {self._metric}"
        if self._distance_map is not None:
            s += ", Cost map: computed "
        else:
            s += ", Cost map: empty "
        if self._valid_plan:
            s += "  \nPlan: valid"
        else:
            s += "  \nPlan: stale"
        return s


    # Just feed self._h into plot function from parent as p

    def next(self, current):
        if not self._valid_plan:
            Error("Cost map has changed, replan")
        x = sub2ind(np.shape(self._costmap), current[1], current[0])
        x = self._b[x]

        n = None
        if x == 0:
            n = np.array([])
        else:
            r, c = ind2sub(np.shape(self._costmap), current[1], current[0])
            n = np.array([c, r])

        return n

    def plan(self, goal=None, animate=False, progress=True):
        if goal is not None:
            self.goal = goal
            self.reset()

        assert(goal is not None and goal is not np.array([]))

        goal = self._goal

        self._g = sub2ind(np.shape(self._occ_grid_nav), goal[1], goal[0])
        self.__insert(self._g, 0, 'goalset')
        self._h[self._g] = 0

        self._niter = 0
        h_prog = None
        if progress:
            h_prog = self.progress_init('D* Planning')

        n_free = np.prod(np.shape(self._occ_grid_nav)) - np.sum(np.sum(self._occ_grid_nav > 0))
        n_update = np.round(n_free / 100)

        while True:
            self._niter = self._niter + 1
            if progress and np.mod(self._niter, n_update) == 0:
                self.progress(h_prog, self._niter/n_free)

                if animate:
                    self.show_distance(self._h)

            if self.process_state() < 0:
                break

            if progress:
                self.progress_delete(h_prog)
            self._valid_plan = True
            disp(self._niter + " iterations\n")

    def set_cost(self, cost_map):
        if not np.all(np.shape(cost_map) == np.shape(self._occ_grid_nav)):
            Error("Costmap must be the same size of the occupancy grid")

        self._costmap = cost_map
        self._valid_plan = False

    def modify_cost(self, xy, new_cost):
        if np.all(np.shape[xy] == np.array([2, 2])) and np.size(new_cost) == 1:
            for xx in range(xy[0, 0], xy[0, 1]):
                for yy in range(xy[1, 0], xy[1, 1]):
                    self.modify(xx, yy, new_cost)
        elif len(xy[0]) == np.size(new_cost):
            for i in range(0, len[xy[0]]):
                self.modify(xy[0, i], xy[1, i], new_cost[i])
        else:
            Error("Number of columns of P and C must match")

        self._valid_plan = False

    def modify(self, x, y, new_cost):
        x = sub2ind(np.shape(self._costmap), y, x)
        self._costmap[x] = new_cost

        if self._t[x] == self._closed:
            self.__insert(x, self._h[x], 'modifycost')

    # These are private methods ... kinda
    def __occ_grid2cost_map(self, og=None, cost=None):
        if cost is None:
            cost = 1
        self._costmap = og
        self._costmap[self._costmap==1] = np.Inf
        self._costmap[self._costmap==0] = cost

    def __process_state(self):
        x = self.__min_state()
        r = None

        if x is None or x is np.array([]):
            r = -1
            return r

        k_old = self.__get_kmin()
        self.__delete(x)

        if k_old < self.h[x]:
            self.message('k_old == h[x]: ' + k_old)
            for y in self.__neighbours(x):
                if (self._h[y] <= k_old) and (self._h[x] > self._h[y] + self.__c(y, x)):
                    self._b[x] = y
                    self._h[x] = self._h[y] + self.__c(y, x)

        if k_old == self._h[x]:
            self.message('k_old == h[x]: ' + k_old)
            for y in self.__neighbours(x):
                if (self._t[y] == self._new) or ( (self._b[y] == x) and
                    (self._h[y] != (self._h[x] + self.c[x, y]))) or  ((self._b[y] != x)
                     and (self._h[y] > (self._h[x] + self.c[x, y]))):

                    self._b[y] = x
                    self.__insert(y, self._h(x)+self._c[x, y], 'L13')
        else:
            self.message("k_old == h[x]: " + k_old)
            for y in self.__neighbours(x):
                if (self._t[y] == self.__new) or ((self._b[y] == x) and
                        (self._h[y] != (self._h[x] + self.__c(x, y)))):
                    self._b[y] = x
                    self.__insert(y, self._h[x] + self.__c(x, y), "L18")
                else:
                    if (self._b[y] != x) and (self._h[y] > (self._h[x] + self.__c(x, y))):
                        self.__insert(x, self._h[x], "L21")
                    else:
                        if (self._b[y] != x) and (self._h[x] > self._h[y] + self.__c(y, x)) and \
                                (self._t[y] == self.__closed) and self._h[y] > k_old:
                            self.__insert(y, self._h[y], 'L25')

        r = 0
        return r

    def __k(self, x):
        i = self._open_list[0,:] == x
        kk = self._open_list[1, i]
        return kk

    def __insert(self, x, h_new, where):
        self.message("Insert (" + where + ") " + x + " = " + h_new + "\n")

        i = np.argwhere(self._open_list[0,:] == x)
        if len(i) > 1:
            Error("D*:Insert: state in open list " + x + " times.\n")

        k_new = None
        if self._t[x] == self._new:
            k_new = h_new
            data = np.array([(x), (k_new)])
            self._open_list = np.array([(self._open_list), data] )
        elif self._t[x] == self._open:
            k_new = np.min(self._open_list[1, i], h_new)
        elif self._t(x) == self._closed:
            k_new = np.min(self._h[x], h_new)
            data = np.array([(x), (k_new)])
            self._open_list = np.array([self._open_list, data])

        if len(self._open_list[0]) > self._open_list_maxlen:
            self._open_list_maxlen = len(self._open_list[0])

        self._h[x] = h_new
        self._t[x] = self._open

    def __delete(self, x):
        self.message("Delete " + x)
        i = np.argwhere(self._open_list[0, :] == x)
        if len(i) != 1:
            Error("D*:Delete: state " + x + " doesn't exist.")
        if len(i) > 1:
            disp("Del")
        self._open_list = np.delete(self._open_list, i-1, 1)
        self._t = self._closed

    def __min_state(self):
        ms = None
        if self._open_list is None or self._open_list is np.array([]):
            ms = np.array([])
        else:
            d, i = np.minimum(self._open_list[1, :])
            ms = self._open_list[0, i]
        return ms

    def __get_kmin(self):
        kmin = np.minimum(self._open_list[1, :])
        return kmin

    def __c(self, x, y):
        r, c = ind2sub(np.shape(self.costmap), np.array([[x], [y]]))
        dist = np.sqrt(np.sum(np.diff(np.square(np.array([r, c])))))
        d_cost = (self._costmap[x] + self._costmap[y])/2

        cost = dist * d_cost
        return cost

    def __neighbours(self, x):
        dims = np.shape(self._costmap)
        r, c = ind2sub(dims, x)

        y = np.array([[r-1, r-1, r-1, r, r,  r+1, r+1, r+1],
                      [c-1, c, c+1, c-1, c+1, c-1, c, c+1]])

        k = (np.minimum(y) > 0) & (y[0, :] <= dims[0]) & (y[1, :] <= dims[1])
        y = y[:, k-1]
        y = np.transpose(sub2ind(dims, np.transpose(y[0, :]), np.transpose(y[1, :])))
        return y


# Sourced from: https://stackoverflow.com/questions/28995146/matlab-ind2sub-equivalent-in-python/28995315#28995315
def sub2ind(array_shape, rows, cols):
    ind = rows*array_shape[1] + cols
    ind[ind < 0] = -1
    ind[ind >= array_shape[0]*array_shape[1]] = -1
    return ind


def ind2sub(array_shape, ind):
    ind[ind < 0] = -1
    ind[ind >= array_shape[0]*array_shape[1]] = -1
    rows = (ind.astype('int') / array_shape[1])
    cols = ind % array_shape[1]
    return rows, cols


if __name__ == "__main__":

    from roboticstoolbox import DStar
    from scipy.io import loadmat

    vars = loadmat("/Users/corkep/code/robotics-toolbox-python/data/map1.mat", squeeze_me=True, struct_as_record=False)
    map = vars['map']
    ds = DStar(map, verbose=True)
    ds = ds.plan([50, 35])
    path = ds.query([20, 10])



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import numpy as np
from collections import namedtuple
from roboticstoolbox.tools.null import null
from spatialmath import base
from spatialmath import SE3, Twist3
import scipy.optimize as opt
import math
import qpsolvers as qp
from roboticstoolbox.tools.p_servo import p_servo
# iksol = namedtuple("IKsolution", "q, success, reason, iterations, residual",
#     defaults=(None, False, None, None, None)) # Py >= 3.7 only
iksol = namedtuple("IKsolution", "q, success, reason, iterations, residual")

# ===================================================================== #


class IKMixin:

    def ikine_mmc(
                self, T,
                q0=None):

        arrived = False

        n = self.n
        q = self.q
        dt = 0.05

        e_prev = 100000
        q_last = q
        gain = 1000.0

        while not arrived:

            Te = self.fkine(q)
            eTep = Te.inv() * T
            e = np.sum(np.abs(np.r_[eTep.t, eTep.rpy() * np.pi/180]))

            if e < e_prev:
                # good update
                # gain = gain * 2
                # dt = dt * 2
                # print('Up')
                pass
            else:
                # bad update
                # gain = gain / 2
                dt = dt / 2
                # q = q_last
                # print('Down')

            e_prev = e
            q_last = q
            # print(gain)
            # print(self.manipulability(q))
            # print(e)

            v, arrived = p_servo(Te, T, gain=gain, threshold=0.000001)

            # Gain term (lambda) for control minimisation
            Y = 0.01

            # Quadratic component of objective function
            Q = np.eye(n + 6)

            # Joint velocity component of Q
            Q[:n, :n] *= Y

            # Slack component of Q
            Q[n:, n:] = (1 / e) * np.eye(6)

            # The equality contraints
            Aeq = np.c_[self.jacobe(q), np.eye(6)]
            beq = v.reshape((6,))

            # The inequality constraints for joint limit avoidance
            Ain = np.zeros((n + 6, n + 6))
            bin = np.zeros(n + 6)

            # The minimum angle (in radians) in which the joint is allowed to
            # approach to its limit
            ps = 0.05

            # The influence angle (in radians) in which the velocity damper
            # becomes active
            pi = 0.9

            Ain[:n, :n], bin[:n] = self.joint_velocity_damper(ps, pi, n)
            c = np.r_[-self.jacobm(q).reshape((n,)), np.zeros(6)]
            # lb = -np.r_[self.qdlim[:n], 10 * np.ones(6)]
            # ub = np.r_[self.qdlim[:n], 10 * np.ones(6)]

            # Solve for the joint velocities dq
            qd = qp.solve_qp(Q, c, Ain, bin, Aeq, beq)

            for i in range(self.n):
                q[i] += qd[i] * (dt)

        return q

# --------------------------------------------------------------------- #

    def ikine_LM(
            self, T,
            q0=None,
            mask=None,
            ilimit=500,
            rlimit=100,
            tol=1e-10,
            L=0.1,
            Lmin=0,
            search=False,
            slimit=100,
            transpose=None):

        """
        Numerical inverse kinematics by Levenberg-Marquadt optimization (Robot superclass)

        :param T: The desired end-effector pose or pose trajectory
        :type T: SE3
        :param q0: initial joint configuration (default all zeros)
        :type q0: ndarray(n)
        :param mask: mask vector that correspond to translation in X, Y and Z
            and rotation about X, Y and Z respectively.
        :type mask: ndarray(6)
        :param ilimit: maximum number of iterations (default 500)
        :type ilimit: int
        :param rlimit: maximum number of consecutive step rejections (default 100)
        :type rlimit: int
        :param tol: final error tolerance (default 1e-10)
        :type tol: float
        :param L: initial value of lambda
        :type L: float (default 0.1)
        :param Lmin: minimum allowable value of lambda
        :type Lmin: float (default 0)
        :param search: search over all configurations
        :type search: bool
        :param slimit: maximum number of search attempts
        :type slimit: int (default 100)
        :param transpose: use Jacobian transpose with step size A, rather
            than Levenberg-Marquadt
        :type transpose: float
        :return: inverse kinematic solution
        :rtype: named tuple

        ``sol = robot.ikine_LM(T)`` are the joint coordinates (n) corresponding
        to the robot end-effector pose ``T`` which is an ``SE3`` object. This method can
        be used for robots with any number of degrees of freedom. The return
        value ``sol`` is a named tuple with elements:

        ============    ==========  ============================================================
        Element         Type        Description
        ============    ==========  ============================================================
        ``q``           ndarray(n)  joint coordinates in units of radians or metres, or ``None``
        ``success``     bool        whether a solution was found
        ``reason``      str         reason for the failure
        ``iterations``  int         number of iterations
        ``residual``    float       final value of cost function
        ============    ==========  ============================================================

        **Trajectory operation**:

        If ``len(T) > 1`` it is considered to be a trajectory, and the result is
        a list of named tuples such that ``sol[k]`` corresponds to ``T[k]``. The
        initial estimate of q for each time step is taken as the solution from
        the previous time step.

        **Underactuated robots:**

        For the case where the manipulator has fewer than 6 DOF the
        solution space has more dimensions than can be spanned by the
        manipulator joint coordinates.

        In this case we specify the ``mask`` option where the ``mask`` vector
        (6) specifies the Cartesian DOF (in the wrist coordinate frame) that
        will be ignored in reaching a solution.  The mask vector has six
        elements that correspond to translation in X, Y and Z, and rotation
        about X, Y and Z respectively. The value should be 0 (for ignore)
        or 1. The number of non-zero elements must equal the number of
        manipulator DOF.

        For example when using a 3 DOF manipulator tool orientation might
        be unimportant, in which case use the option ``mask=[1 1 1 0 0 0]``.

        **Global search**:

        ``sol = robot.ikine_LM(T, search=True)`` as above but peforms a
        brute-force search with initial conditions chosen randomly from the
        entire configuration space.  If a numerical solution is found from that
        initial condition, it is returned, otherwise another initial condition is
        chosen.

        .. note::

            - See `Toolbox kinematics wiki page <https://github.com/petercorke/robotics-toolbox-python/wiki/Kinematics>`_
            - Implements a Levenberg-Marquadt variable-step-size solver.
            - The tolerance is computed on the norm of the error between
              current and desired tool pose.  This norm is computed from
              distances and angles without any kind of weighting.
            - The inverse kinematic solution is generally not unique, and
              depends on the initial guess ``q0``.
            - The default value of ``q0`` is zero which is a poor choice for 
              most manipulators since it often corresponds to a
              kinematic singularity.
            - Such a solution is completely general, though much less
              efficient than analytic inverse kinematic solutions derived
              symbolically.
            - This approach allows a solution to be obtained at a singularity,
              but the joint angles within the null space are arbitrarily
              assigned.
            - Joint offsets, if defined, are accounted for in the solution.
            - Joint limits are not considered in this solution.
            - If the search option is used any prismatic joint must have
              joint limits defined.

        :references:
            - Robotics, Vision & Control, P. Corke, Springer 2011,
              Section 8.4.

        :seealso: :func:`ikine_LMS`, :func:`ikine_unc`, :func:`ikine_con`, :func:`ikine_min`
        """

        if not isinstance(T, SE3):
            T = SE3(T)

        solutions = []

        if search:
            # Randomised search for a starting point
            # quiet = True

            qlim = self.qlim
            qspan = qlim[1] - qlim[0]  # range of joint motion

            for k in range(slimit):
                # choose a random joint coordinate
                q0_k = np.random.rand(self.n) * qspan + qlim[0, :]

                # recurse into the solver
                solution = self.ikine_LM(
                    T[0],
                    q0_k,
                    mask,
                    ilimit,
                    rlimit,
                    tol,
                    L,
                    Lmin,
                    False,
                    slimit,
                    transpose)

                if solution.success:
                    q0 = solution.q
                    if len(T) == 1:
                        # we're done
                        return solution
                    else:
                        # more to do on the trajectory
                        solutions.append(solution)
                        del T[0]
            else:
                # no solution found, stop now
                return iksol(None, False, None, None, None)

        if q0 is None:
            q0 = np.zeros((self.n,))
        else:
            q0 = base.getvector(q0, self.n)

        if mask is not None:
            mask = base.getvector(mask, 6)
            if not self.n >= np.sum(mask):
                raise ValueError('Number of robot DOF must be >= the number '
                                 'of 1s in the mask matrix')
        else:
            mask = np.ones(6)
        W = np.diag(mask)

        tcount = 0    # Total iteration count
        rejcount = 0  # Rejected step count
        nm = 0

        # bool vector indicating revolute joints
        revolutes = np.array([link.isrevolute for link in self])

        q = q0
        for Tk in T:
            iterations = 0
            Li = L  # lambda
            failure = None
            while True:
                # Update the count and test against iteration limit
                iterations += 1

                if iterations > ilimit:
                    failure = f"iteration limit {ilimit} exceeded"
                    break

                e = base.tr2delta(self.fkine(q).A, Tk.A)

                # Are we there yet?
                if base.norm(W @ e) < tol:
                    break

                # Compute the Jacobian
                J = self.jacobe(q)

                JtJ = J.T @ W @ J

                if transpose is not None:
                    # Do the simple Jacobian transpose with constant gain
                    dq = transpose * J.T @ e    # lgtm [py/multiple-definition]
                    q += dq
                else:
                    # Do the damped inverse Gauss-Newton with
                    # Levenberg-Marquadt
                    # dq = np.linalg.inv(
                    #     JtJ + ((Li + Lmin) * np.eye(self.n))
                    # ) @ J.T @ W @ e
                    dq = np.linalg.inv(
                        JtJ + ((Li + Lmin) * np.diag(np.diag(JtJ)))
                    ) @ J.T @ W @ e
                    # print(J.T @ W @ e)

                    # Compute possible new value of
                    qnew = q + dq

                    # And figure out the new error
                    enew = base.tr2delta(self.fkine(qnew).A, Tk.A)

                    # Was it a good update?
                    if np.linalg.norm(W @ enew) < np.linalg.norm(W @ e):
                        # Step is accepted
                        q = qnew
                        e = enew
                        Li /= 2
                        rejcount = 0
                    else:
                        # Step is rejected, increase the damping and retry
                        Li *= 2
                        rejcount += 1
                        if rejcount > rlimit:
                            failure = f"rejected-step limit {rlimit} exceeded"
                            break

                # Wrap angles for revolute joints
                k = np.logical_and(q > np.pi, revolutes)
                q[k] -= 2 * np.pi

                k = np.logical_and(q < -np.pi, revolutes)
                q[k] += + 2 * np.pi

                nm = np.linalg.norm(W @ e)
                qs = ", ".join(["{:8.3f}".format(qi) for qi in q])
                # print(f"λ={Li:8.2g}, |e|={nm:8.2g}: q={qs}")

            # LM process finished, for better or worse
            # failure will be None or an error message
            solution = iksol(q, failure is None, failure, iterations, nm)
            solutions.append(solution)

            tcount += iterations

        if len(T) == 1:
            return solutions[0]
        else:
            return solutions

# --------------------------------------------------------------------- #

    def ikine_LMS(
            self, T,
            q0=None,
            mask=None,
            ilimit=500,
            tol=1e-10,
            wN=1e-3,
            Lmin=0
            ):

        """
        Numerical inverse kinematics by Levenberg-Marquadt optimization (Robot superclass)

        :param T: The desired end-effector pose or pose trajectory
        :type T: SE3
        :param q0: initial joint configuration (default all zeros)
        :type q0: ndarray(n)
        :param mask: mask vector that correspond to translation in X, Y and Z
            and rotation about X, Y and Z respectively.
        :type mask: ndarray(6)
        :param ilimit: maximum number of iterations (default 500)
        :type ilimit: int 
        :param tol: final error tolerance (default 1e-10)
        :type tol: float 
        :param ωN: damping coefficient
        :type ωN: float (default 1e-3)
        :return: inverse kinematic solution
        :rtype: named tuple

        ``sol = robot.ikine_LM(T)`` are the joint coordinates (n) corresponding
        to the robot end-effector pose ``T`` which is an ``SE3`` object. This method can
        be used for robots with any number of degrees of freedom. The return
        value ``sol`` is a named tuple with elements:

        ============    ==========  ============================================================
        Element         Type        Description
        ============    ==========  ============================================================
        ``q``           ndarray(n)  joint coordinates in units of radians or metres, or ``None``
        ``success``     bool        whether a solution was found
        ``reason``      str         reason for the failure
        ``iterations``  int         number of iterations
        ``residual``    float       final value of cost function
        ============    ==========  ============================================================

        **Trajectory operation**:

        If ``len(T) > 1`` it is considered to be a trajectory, and the result is
        a list of named tuples such that ``sol[k]`` corresponds to ``T[k]``. The
        initial estimate of q for each time step is taken as the solution from
        the previous time step.

        **Underactuated robots:**

        For the case where the manipulator has fewer than 6 DOF the
        solution space has more dimensions than can be spanned by the
        manipulator joint coordinates.

        In this case we specify the ``mask`` option where the ``mask`` vector
        (6) specifies the Cartesian DOF (in the wrist coordinate frame) that
        will be ignored in reaching a solution.  The mask vector has six
        elements that correspond to translation in X, Y and Z, and rotation
        about X, Y and Z respectively. The value should be 0 (for ignore)
        or 1. The number of non-zero elements should equal the number of
        manipulator DOF.

        For example when using a 3 DOF manipulator rotation orientation might
        be unimportant in which case use the option: mask = [1 1 1 0 0 0].

        .. note::

            - See `Toolbox kinematics wiki page <https://github.com/petercorke/robotics-toolbox-python/wiki/Kinematics>`_
            - Implements a modified Levenberg-Marquadt variable-step-size solver
              which is quite robust in practice.
            - The tolerance is computed on the norm of the error between
              current and desired tool pose.  This norm is computed from
              distances and angles without any kind of weighting.
            - The inverse kinematic solution is generally not unique, and
              depends on the initial guess ``q0``.
            - The default value of ``q0`` is zero which is a poor choice for 
              most manipulators since it often corresponds to a
              kinematic singularity.
            - Such a solution is completely general, though much less
              efficient than analytic inverse kinematic solutions derived
              symbolically.
            - This approach allows a solution to be obtained at a singularity,
              but the joint angles within the null space are arbitrarily
              assigned.
            - Joint offsets, if defined, are accounted for in the solution.
            - Joint limits are not considered in this solution.

        :references:
            - "Solvability-Unconcerned Inverse Kinematics by the
              Levenberg–Marquardt Method", T. Sugihara, IEEE T-RO, 27(5), 
              October 2011, pp. 984-991.

        :seealso: :func:`ikine_LM`, :func:`ikine_unc`, :func:`ikine_con`, :func:`ikine_min`
        """

        if not isinstance(T, SE3):
            T = SE3(T)

        solutions = []

        if q0 is None:
            q0 = np.zeros((self.n,))
        else:
            q0 = base.getvector(q0, self.n)

        if mask is not None:
            mask = base.getvector(mask, 6)
            if not self.n >= np.sum(mask):
                raise ValueError('Number of robot DOF must be >= the number '
                                'of 1s in the mask matrix')
        else:
            mask = np.ones(6)
        W = np.diag(mask)

        tcount = 0  # Total iteration count

        # bool vector indicating revolute joints
        revolutes = np.array([link.isrevolute for link in self])

        q = q0
        for Tk in T:
            iterations = 0
            failure = None
            while True:
                # Update the count and test against iteration limit
                iterations += 1

                if iterations > ilimit:
                    failure = f"iteration limit {ilimit} exceeded"
                    break

                e = _angle_axis(self.fkine(q).A, Tk.A)

                # Are we there yet?
                E = 0.5 * e.T @ W @ e
                if E < tol:
                    break

                # Compute the Jacobian and projection matrices
                J = self.jacob0(q)
                WN = E * np.eye(self.n) + wN * np.eye(self.n)
                H = J.T @ W @ J + WN  # n x n
                g = J.T @ W @ e       # n x 1

                # Compute new value of q
                q += np.linalg.inv(H) @ g  # n x 1
                # print(np.linalg.norm(np.linalg.inv(H) @ g))
                # print(e)
                # print(g)
                # print(q)
                # print(J)

                # Wrap angles for revolute joints
                k = np.logical_and(q > np.pi, revolutes)
                q[k] -= 2 * np.pi

                k = np.logical_and(q < -np.pi, revolutes)
                q[k] += + 2 * np.pi

                qs = ", ".join(["{:8.3f}".format(qi) for qi in q])
                # print(f"|e|={E:8.2g}, det(H)={np.linalg.det(H)}: q={qs}")

            # LM process finished, for better or worse
            # failure will be None or an error message
            solution = iksol(q, failure is None, failure, iterations, E)
            solutions.append(solution)

            tcount += iterations

        if len(T) == 1:
            return solutions[0]
        else:
            return solutions


# --------------------------------------------------------------------- #

    # def ikine_unc(self, T, q0=None, ilimit=1000, tol=1e-16, stiffness=0, costfun=None):
    #     r"""
    #     Inverse manipulator by optimization without joint limits (Robot
    #     superclass)

    #     :param T: The desired end-effector pose or pose trajectory
    #     :type T: SE3
    #     :param q0: initial joint configuration (default all zeros)
    #     :type q0: ndarray(n)
    #     :param tol: Tolerance (default 1e-16)
    #     :type tol: tol
    #     :param ilimit: Iteration limit (default 1000)
    #     :type ilimit: int
    #     :param stiffness: Stiffness used to impose a smoothness contraint on
    #         joint angles, useful when n is large (default 0)
    #     :type stiffness: float
    #     :param costfun: User supplied cost term, optional
    #     :type costfun: callable
    #     :return: inverse kinematic solution
    #     :rtype: named tuple

    #     ``sol = robot.ikine_unc(T)`` are the joint coordinates (n) corresponding
    #     to the robot end-effector pose T which is an SE3 object.  The
    #     return value ``sol`` is a named tuple with elements:

    #     ============    ==========  ============================================================
    #     Element         Type        Description
    #     ============    ==========  ============================================================
    #     ``q``           ndarray(n)  joint coordinates in units of radians or metres, or ``None``
    #     ``success``     bool        whether a solution was found
    #     ``reason``      str         reason for the failure
    #     ``iterations``  int         number of iterations
    #     ``residual``    float       final value of cost function
    #     ============    ==========  ============================================================

    #     This method the Scipy SLSQP minimizer to minimize the squared norm of a
    #     vector :math:`[d,a]` with components respectively the translation error
    #     and rotation error in Euler vector form, between the desired pose and
    #     the current estimate obtained by inverse kinematics.

    #     **Additional cost terms**:

    #     This method supports two additional costs:

    #     - ``stiffness`` imposes a penalty on joint variation :math:`\sum_{j=1}^N (q_j - q_{j-1})^2`
    #       which tends to keep the arm straight
    #     - ``costfun`` add a cost given by a user-specified function ``costfun(q)``

    #     **Trajectory operation**:

    #     If ``len(T) > 1`` it is considered to be a trajectory, and the result is
    #     a list of named tuples such that ``sol[k]`` corresponds to ``T[k]``. The
    #     initial estimate of q for each time step is taken as the solution from
    #     the previous time step.


    #     .. note::

    #         - Uses ``SciPy.minimize`` SLSQP without bounds.
    #         - Joint limits are not considered in this solution.
    #         - Can be used for robots with arbitrary degrees of freedom.
    #         - The inverse kinematic solution is generally not unique, and
    #           depends on the initial guess ``q0``.
    #         - The default value of ``q0`` is zero which is a poor choice for 
    #           most manipulators since it often corresponds to a
    #           kinematic singularity.
    #         - Such a solution is completely general, though much less
    #           efficient than analytic inverse kinematic solutions derived
    #           symbolically.
    #         - The objective function (error) is 
    #           :math:`\sum \left( (\mat{T}^{-1} \cal{K}(\vec{q}) - \mat{1} ) \mat{\Omega} \right)^2`
    #           where :math:`\mat{\Omega}` is a diagonal matrix.
    #         - Joint offsets, if defined, are accounted for in the solution.

    #     .. warning:: 
        
    #         - The objective function is rather uncommon.
    #         - Order of magnitude slower than ``ikine_LM`` or ``ikine_LMS``, it
    #           uses a scalar cost-function and does not provide a Jacobian.

    #     :seealso: :func:`ikine_LM`, :func:`ikine_LMS`, :func:`ikine_con`, :func:`ikine_min`
    #     """

    #     if not isinstance(T, SE3):
    #         T = SE3(T)

    #     if q0 is None:
    #         q0 = np.zeros((self.n))
    #     else:
    #         q0 = base.getvector(q0, self.n)

    #     solutions = []

    #     wr = 1 / self.reach
    #     weight = np.r_[wr, wr, wr, 1, 1, 1]

    #     def cost(q, T, weight, costfun, stiffness):
    #         # T, weight, costfun, stiffness = args
    #         e = _angle_axis(self.fkine(q).A, T) * weight
    #         E = (e**2).sum()

    #         if stiffness > 0:
    #             # Enforce a continuity constraint on joints, minimum bend
    #             E += np.sum(np.diff(q)**2) * stiffness

    #         if costfun is not None:
    #             E += (e**2).sum() + costfun(q)

    #         return E

    #     for Tk in T:

    #         res = minimize(
    #             cost,
    #             q0,
    #             args=(Tk.A, weight, costfun, stiffness),
    #             tol=tol,
    #             method='SLSQP',
    #             options={'maxiter': ilimit}
    #             )

    #             # final gradient tolerance must be < gtol for success, bump
    #             # this number up a bit
    #             # SLSQP seems to work better than BFGS, L-BFGS-B

    #         solution = iksol(res.x, res.success, res.message, res.nit, res.fun)
    #         solutions.append(solution)
    #         q0 = res.x  # use this solution as initial estimate for next time

    #     if len(T) == 1:
    #         return solutions[0]
    #     else:
    #         return solutions

# --------------------------------------------------------------------- #

    def ikine_min(self, T, q0=None, qlim=False, ilimit=1000, tol=1e-16, method=None, stiffness=0, costfun=None, options={}):
        r"""
        Inverse kinematics by optimization with joint limits (Robot superclass)

        :param T: The desired end-effector pose or pose trajectory
        :type T: SE3
        :param q0: initial joint configuration (default all zeros)
        :type q0: ndarray(n)
        :param qlim: enforce joint limits
        :type qlim: bool
        :param ilimit: Iteration limit (default 1000)
        :type ilimit: int
        :param tol: Tolerance (default 1e-16)
        :type tol: tol
        :param method: minimization method to use
        :type method: str
        :param stiffness: Stiffness used to impose a smoothness contraint on
            joint angles, useful when n is large (default 0)
        :type stiffness: float
        :param costfun: User supplied cost term, optional
        :type costfun: callable
        :return: inverse kinematic solution
        :rtype: named tuple

        ``sol = robot.ikine_min(T)`` are the joint coordinates (n) corresponding
        to the robot end-effector pose T which is an SE3 object.  The
        return value ``sol`` is a named tuple with elements:

        ============    ==========  ============================================================
        Element         Type        Description
        ============    ==========  ============================================================
        ``q``           ndarray(n)  joint coordinates in units of radians or metres, or ``None``
        ``success``     bool        whether a solution was found
        ``reason``      str         reason for the failure
        ``iterations``  int         number of iterations
        ``residual``    float       final value of cost function
        ============    ==========  ============================================================

        **Minimization method**:

        By default this method uses:

        - the Scipy ``SLSQP`` (Sequential Least Squares Programming) minimizer
          for the case of no joint limits
        - the Scipy ``trust-constr`` minimizer for the case with joint limits.
          This gives good results but is very slow.  An alternative is
          ``L-BFGS-B`` (Broyden–Fletcher–Goldfarb–Shanno) but for redundant
          robots can sometimes give poor results, pushing against the joint
          limits when there is no need to.
          
        In both case the function to be minimized is the squared norm of a 
        vector :math:`[d,a]` with components respectively the
        translation error and rotation error in Euler vector form, between the
        desired pose and the current estimate obtained by inverse kinematics.

        **Additional cost terms**:

        This method supports two additional costs:

        - ``stiffness`` imposes a penalty on joint variation 
          :math:`\sum_{j=1}^N (q_j - q_{j-1})^2` which tends to keep the 
          arm straight
        - ``costfun`` add a cost given by a user-specified function ``costfun(q)``

        **Trajectory operation**:

        If ``len(T) > 1`` it is considered to be a trajectory, and the result is
        a list of named tuples such that ``sol[k]`` corresponds to ``T[k]``. The
        initial estimate of q for each time step is taken as the solution from
        the previous time step.

        .. note::

            - See `Toolbox kinematics wiki page <https://github.com/petercorke/robotics-toolbox-python/wiki/Kinematics>`_
            - Uses ``SciPy.minimize`` with bounds.
            - Joint limits are considered in this solution.
            - Can be used for robots with arbitrary degrees of freedom.
            - The inverse kinematic solution is generally not unique, and
              depends on the initial guess ``q0``.
            - The default value of ``q0`` is zero which is a poor choice for 
              most manipulators since it often corresponds to a
              kinematic singularity.
            - Such a solution is completely general, though much less
              efficient than analytic inverse kinematic solutions derived
              symbolically.
            - The objective function (error) is 
              :math:`\sum \left( (\mat{T}^{-1} \cal{K}(\vec{q}) - \mat{1} ) \mat{\Omega} \right)^2`
              where :math:`\mat{\Omega}` is a diagonal matrix.
            - Joint offsets, if defined, are accounted for in the solution.

        .. warning:: 
        
            - The objective function is rather uncommon.
            - Order of magnitude slower than ``ikine_LM`` or ``ikine_LMS``, it
              uses a scalar cost-function and does not provide a Jacobian.

        :author: Bryan Moutrie, for RTB-MATLAB

        :seealso: :func:`ikine_LM`, :func:`ikine_LMS`, :func:`ikine_unc`, :func:`ikine_min`

        """

        if not isinstance(T, SE3):
            T = SE3(T)

        if q0 is None:
            q0 = np.zeros((self.n))
        else:
            q0 = base.getvector(q0, self.n)

        solutions = []

        wr = 1 / self.reach
        weight = np.r_[wr, wr, wr, 1, 1, 1]

        optdict = {'maxiter': ilimit}
        if options is not None and isinstance(options, dict):
            optdict.update(options)
        else:
            raise ValueError('options must be a dict')

        if qlim:
            # dealing with joint limits
            bounds = opt.Bounds(self.qlim[0, :], self.qlim[1, :])

            if method is None:
                method='trust-constr'
        else:
            # no joint limits
            if method is None:
                method = 'SLSQP'
            bounds = None

        def cost(q, T, weight, costfun, stiffness):
            # T, weight, costfun, stiffness = args
            e = _angle_axis(self.fkine(q).A, T) * weight
            E = (e**2).sum()

            if stiffness > 0:
                # Enforce a continuity constraint on joints, minimum bend
                E += np.sum(np.diff(q)**2) * stiffness

            if costfun is not None:
                E += (e**2).sum() + costfun(q)

            return E

        for Tk in T:
            res = opt.minimize(
                cost,
                q0, 
                args=(Tk.A, weight, costfun, stiffness),
                bounds=bounds,
                method=method,
                tol=tol,
                options=options
            )

            # trust-constr seems to work better than L-BFGS-B which often
            # runs a joint up against its limit and terminates with position 
            # error.
            # but 'truts-constr' is 5x slower

            solution = iksol(res.x, res.success, res.message, res.nit, res.fun)
            solutions.append(solution)
            q0 = res.x  # use this solution as initial estimate for next time

        if len(T) == 1:
            return solutions[0]
        else:
            return solutions

# --------------------------------------------------------------------- #

    def ikine_global(self, T, q0=None, qlim=False, ilimit=1000, tol=1e-16, method=None, options={}):
        r"""
        .. warning:: Experimental code for using SciPy global optimizers.

        Each global optimizer has quite a different call signature, so final
        design will need a bit of thought.

        """

        # basinhopping:
        # brute: ranges, finish=None
        # differential_evolution:  bounds, tol
        # shgo: bounds, options:f_tol
        # dual_annealing: bounds

        if not isinstance(T, SE3):
            T = SE3(T)

        if q0 is None:
            q0 = np.zeros((self.n))
        else:
            q0 = base.getvector(q0, self.n)

        solutions = []

        wr = 1 / self.reach
        weight = np.r_[wr, wr, wr, 1, 1, 1]

        optdict = {}

        if method is None:
            method='differential-evolution'

        if method == 'brute':
            # requires a tuple of tuples
            optdict['ranges'] = tuple([tuple(l.qlim) for l in self])
        else:
            optdict['bounds'] = tuple([tuple(l.qlim) for l in self])


        if method not in ['basinhopping', 'brute', 'differential_evolution',
                          'shgo', 'dual_annealing']:
            raise ValueError('unknown global optimizer requested')

        global_minimizer = opt.__dict__[method]

        def cost(q, T, weight):
            # T, weight, costfun, stiffness = args
            e = _angle_axis(self.fkine(q).A, T) * weight
            return (e**2).sum()

        for Tk in T:
            res = global_minimizer(
                cost,
                **optdict)
            

            solution = iksol(res.x, res.success, res.message, res.nit, res.fun)
            solutions.append(solution)
            q0 = res.x  # use this solution as initial estimate for next time

        if len(T) == 1:
            return solutions[0]
        else:
            return solutions
            
# --------------------------------------------------------------------- #

    # def ikine_min(self, T, q0=None, pweight=1.0, stiffness=0.0,
    #            qlimits=True, ilimit=1000):
    #     """
    #     Inverse kinematics by optimization with joint limits (Robot superclass)

    #     :param T: The desired end-effector pose or pose trajectory
    #     :type T: SE3
    #     :param q0: initial joint configuration (default all zeros)
    #     :type q0: ndarray(n)
    #     :param pweight: weighting on position error norm compared to rotation
    #         error (default 1)
    #     :type pweight: float
    #     :param stiffness: Stiffness used to impose a smoothness contraint on
    #         joint angles, useful when n is large (default 0)
    #     :type stiffness: float
    #     :param qlimits: Enforce joint limits (default True)
    #     :type qlimits: bool
    #     :param ilimit: Iteration limit (default 1000)
    #     :type ilimit: bool
    #     :return: inverse kinematic solution
    #     :rtype: named tuple

    #     ``sol = robot.ikine_unc(T)`` are the joint coordinates (n) corresponding
    #     to the robot end-effector pose T which is an SE3 object.  The
    #     return value ``sol`` is a named tuple with elements:

    #     ============    ==========  ============================================================
    #     Element         Type        Description
    #     ============    ==========  ============================================================
    #     ``q``           ndarray(n)  joint coordinates in units of radians or metres, or ``None``
    #     ``success``     bool        whether a solution was found
    #     ``reason``      str         reason for the failure
    #     ``iterations``  int         number of iterations
    #     ``residual``    float       final value of cost function
    #     ============    ==========  ============================================================

    #     **Trajectory operation**:

    #     If ``len(T) > 1`` it is considered to be a trajectory, and the result is
    #     a list of named tuples such that ``sol[k]`` corresponds to ``T[k]``. The
    #     initial estimate of q for each time step is taken as the solution from
    #     the previous time step.

    #     .. note::

    #         - PROTOTYPE CODE UNDER DEVELOPMENT, intended to do numerical
    #           inverse kinematics with joint limits
    #         - Uses ``SciPy.minimize`` with/without constraints.
    #         - The inverse kinematic solution is generally not unique, and
    #           depends on the initial guess ``q0``.
    #         - This norm is computed from distances and angles and ``pweight``
    #           can be used to scale the position error norm to be congruent 
    #           with rotation error norm.
    #         - For a highly redundant robot ``stiffness`` can be used to impose 
    #           a smoothness contraint on joint angles, tending toward solutions
    #           with are smooth curves.
    #         - The default value of ``q0`` is zero which is a poor choice for 
    #           most manipulators since it often corresponds to a
    #           kinematic singularity.
    #         - Such a solution is completely general, though much less
    #           efficient than analytic inverse kinematic solutions derived
    #           symbolically.
    #         - This approach allows a solution to obtained at a singularity,
    #           but the joint angles within the null space are arbitrarily
    #           assigned.
    #         - Joint offsets, if defined, are accounted for in the solution.
    #         - Joint limits become explicit bounds if 'qlimits' is set.

    #     .. warning:: 
        
    #         - The objective function is rather uncommon.
    #         - Order of magnitude slower than ``ikine_LM`` or ``ikine_LMS``, it
    #           uses a scalar cost-function and does not provide a Jacobian.

    #     :seealso: :func:`ikine_LM`, :func:`ikine_LMS`, :func:`ikine_unc`, :func:`ikine_con`
    #     """

    #     if not isinstance(T, SE3):
    #         T = SE3(T)

    #     if q0 is None:
    #         q0 = np.zeros((self.n,))
    #     else:
    #         q0 = base.getvector(q0, self.n)

    #     col = 2
    #     solutions = []

    #     # Define the cost function to minimise
    #     def cost(q, *args):
    #         T, pweight, col, stiffness = args
    #         Tq = self.fkine(q).A

    #         # translation error
    #         dT = base.transl(T) - base.transl(Tq)
    #         E = np.linalg.norm(dT) * pweight

    #         # Rotation error
    #         # Find dot product of two columns
    #         dd = np.dot(T[0:3, col], Tq[0:3, col])
    #         E += np.arccos(dd)**2 * 1000

    #         if stiffness > 0:
    #             # Enforce a continuity constraint on joints, minimum bend
    #             E += np.sum(np.diff(q)**2) * stiffness

    #         return E

    #     for Tk in T:

    #         if qlimits:
    #             bounds = Bounds(self.qlim[0, :], self.qlim[1, :])

    #             res = minimize(
    #                 cost,
    #                 q0, 
    #                 args=(Tk.A, pweight, col, stiffness),
    #                 bounds=bounds,
    #                 options={'gtol': 1e-6, 'maxiter': ilimit})
    #         else:
    #             # No joint limits, unconstrained optimization
    #             # final gradient tolerance must be < gtol for success, bump
    #             # this number up a bit
    #             res = minimize(
    #                 cost,
    #                 q0, 
    #                 args=(Tk.A, pweight, col, stiffness),
    #                 options={'gtol': 1e-6, 'maxiter': ilimit})

    #         solution = iksol(res.x, res.success, res.message, res.nit, res.fun)
    #         solutions.append(solution)
    #         q0 = res.x  # use this solution as initial estimate for next time

    #     if len(T) == 1:
    #         return solutions[0]
    #     else:
    #         return solutions


    # def qmincon(self, q=None):
    #     """
    #     Move away from joint limits

    #     :param q: Joint coordinates
    #     :type q: ndarray(n)
    #     :retrun qs: The calculated joint values
    #     :rtype qs: ndarray(n)
    #     :return: Optimisation solved (True) or failed (False)
    #     :rtype: bool
    #     :return: Final value of the objective function
    #     :rtype: float

    #     ``qs, success, err = qmincon(q)`` exploits null space motion and
    #     returns a set of joint angles ``qs`` (n) that result in the same
    #     end-effector pose but are away from the joint coordinate limits.
    #     ``n`` is the number of robot joints. ``success`` is True for
    #     successful optimisation. ``err`` is the scalar final value of
    #     the objective function.

    #     **Trajectory operation**

    #     In all cases if ``q`` is (m,n) it is taken as a pose sequence and
    #     ``qmincon()`` returns the adjusted joint coordinates (m,n)
    #     corresponding to each of the configurations in the sequence.

    #     ``err`` and ``success`` are also (m) and indicate the results of
    #     optimisation for the corresponding trajectory step.

    #     .. note:: Robot must be redundant.

    #     """

    #     def sumsqr(A):
    #         return np.sum(A**2)

    #     def cost(x, ub, lb, qm, N):
    #         return sumsqr(
    #             (2 * (N @ x + qm) - ub - lb) / (ub - lb))

    #     q = getmatrix(q, (None, self.n))

    #     qstar = np.zeros((q.shape[0], self.n))
    #     error = np.zeros(q.shape[0])
    #     success = np.zeros(q.shape[0])

    #     lb = self.qlim[0, :]
    #     ub = self.qlim[1, :]

    #     for k, qk in enumerate(q):

    #         J = self.jacobe(qk)

    #         N = null(J)

    #         x0 = np.zeros(N.shape[1])
    #         A = np.r_[N, -N]
    #         b = np.r_[ub - qk, qk - lb].reshape(A.shape[0],)

    #         con = LinearConstraint(A, -np.inf, b)

    #         res = minimize(
    #             lambda x: cost(x, ub, lb, qk, N),
    #             x0, constraints=con)

    #         qstar[k, :] = qk + N @ res.x
    #         error[k] = res.fun
    #         success[k] = res.success

    #     if q.shape[0] == 1:
    #         return qstar[0, :], success[0], error[0]
    #     else:
    #         return qstar, success, error

def _angle_axis(T, Td):
    d = base.transl(Td) - base.transl(T)
    R = base.t2r(Td) @ base.t2r(T).T
    l = np.r_[R[2,1]-R[1,2], R[0,2]-R[2,0], R[1,0]-R[0,1]]
    if base.iszerovec(l):
        # diagonal matrix case
        if np.trace(R) > 0:
            # (1,1,1) case
            a = np.zeros((3,))
        else:
            a = np.pi / 2 * (np.diag(R) + 1)
    else:
        # non-diagonal matrix case
        ln = base.norm(l)
        a = math.atan2(ln, np.trace(R) - 1) * l / ln
        
    return np.r_[d, a]

def _angle_axis_sekiguchi(T, Td):
    d = base.transl(Td) - base.transl(T)
    R = base.t2r(Td) @ base.t2r(T).T
    l = np.r_[R[2,1]-R[1,2], R[0,2]-R[2,0], R[1,0]-R[0,1]]
    if base.iszerovec(l):
        # diagonal matrix case
        if np.trace(R) > 0:
            # (1,1,1) case
            a = np.zeros((3,))
        else:
            # (1, -1, -1), (-1, 1, -1) or (-1, -1, 1) case
            a = np.pi / 2 * (np.diag(R) + 1)
            # as per Sekiguchi paper
            if R[1,0] > 0 and R[2,1] > 0 and R[0,2] > 0:
                a = np.pi / np.sqrt(2) * np.sqrt(n.diag(R) + 1)
            elif R[1,0] > 0: # (2)
                a = np.pi / np.sqrt(2) * np.sqrt(n.diag(R) @ np.r_[1,1,-1] + 1)
            elif R[0,2] > 0: # (3)
                a = np.pi / np.sqrt(2) * np.sqrt(n.diag(R) @ np.r_[1,-1,1] + 1)
            elif R[2,1] > 0: # (4)
                a = np.pi / np.sqrt(2) * np.sqrt(n.diag(R) @ np.r_[-1,1,1] + 1)
    else:
        # non-diagonal matrix case
        ln = base.norm(l)
        a = math.atan2(ln, np.trace(R) - 1) * l / ln

    return np.r_[d, a]


if __name__ == "__main__":  # pragma nocover

    import roboticstoolbox as rtb
    from spatialmath import SE3

    # np.set_printoptions(linewidth=120, formatter={'float': lambda x: f"{x:9.5g}" if abs(x) > 1e-10 else f"{0:9.5g}"})


    robot = rtb.models.DH.Panda()

    T = SE3(0.7, 0.2, 0.1) * SE3.OA([0, 1, 0], [0, 0, -1])
    # sol = robot.ikine_LMS(T)         # solve IK
    # print(sol)                    # display joint angles

    # print(T)
    # print(robot.fkine(sol.q))
    # robot.plot(sol.q)

    # sol = robot.ikine_unc(T, costfun=lambda q: q[1] * 1e-6 if q[1] > 0 else -q[1])
    # print(sol)
    # print(robot.fkine(sol.q))
    # robot.plot(sol.q)

    sol = robot.ikine_global(T, method='brute')


"""
Python ReedShepp Planner
@Author: Kristian Gibson
TODO: Comments + Sphynx Docs Structured Text
TODO: Bug-fix, testing

Not ready for use yet.
"""
from numpy import disp
from scipy import integrate
from spatialmath.base.transforms2d import *
from spatialmath.base.vectors import *
from spatialmath.pose2d import SE2
from spatialmath import base
from scipy.ndimage import *
from matplotlib import cm
from roboticstoolbox.mobile.navigation import Navigation


class DXform(Navigation):
    def __init__(self, occ_grid=None, metric="euclidean", distance_map=None, **kwargs):

        super().__init__(occ_grid=occ_grid, **kwargs)
        self._metric = metric
        self._distance_map = None

    @property
    def metric(self):
        return self._metric

    @property
    def distance_map(self):
        return self._distance_map

    def __str__(self):
        s = super().__str__()
        s += f"\n  Distance metric: {self._metric}"
        if self._distance_map is not None:
            s += ", Distance map: computed "
        else:
            s += ", Distance map: empty "

        return s

    def goal_change(self, goal):
        self._distance_map = np.array([])

    def plan(self, goal=None, animate=False):
        show = None
        if animate:
            show = 0.05
        else:
            show = 0

        if goal is not None:
            self.goal = goal

        if self._goal is None:
            raise ValueError('No goal specified here or in constructor')

        self._distance_map = distancexform(self.occ_grid_nav, self._goal, self._metric)

    # Use plot from parent class

    def next(self, robot):
        if self._distance_map is None:
            Error("No distance map computed, you need to plan.")

        directions = np.array([
            # dy  dx
            [-1, -1],
            [ 0, -1],
            [ 1, -1],
            [-1,  0],
            [ 0,  0],
            [ 1,  0],
            [ 0,  1],
            [ 1,  1],
        ])

        x = robot[0]
        y = robot[1]

        min_dist = np.inf
        for d in directions:
            try:
                if self._distance_map[y + d[0], x + d[1]] < min_dist:
                    min_dir = d
                    min_dist = self.distance_map[y + d[0], x + d[1]]
            except:
                # come here if the neighbouring cell is outside the map bounds
                raise RuntimeError(f"Unexpected error finding next min dist at {d}")

        if np.isinf(min_dist):
            raise RuntimeError("no minimum found, shouldn't happen")

        x = x + min_dir[1]
        y = y + min_dir[0]

        next = np.r_[x, y]
        if all(next == self._goal):
            return None
        else:
            return next

    def plot_3d(self, p=None, ls=None):
        fig = plt.figure()
        ax = fig.gca(projection='3d')

        surf = ax.plot_surface(self._distance_map, cmap=cm.coolwarm,
                               linewidth=0, antialiased=False)

        if p is not None:
            k = sub2ind(np.shape(self._distance_map), p[:, 1], p[:, 0])
            height = self._distance_map[k]
            ax.plot(p[:, 0], p[:, 1], height)

        plt.show()

    def plot(self, **kwargs):
        super().plot(distance=self._distance_map, **kwargs)

# Sourced from: https://stackoverflow.com/questions/28995146/matlab-ind2sub-equivalent-in-python/28995315#28995315
def sub2ind(array_shape, rows, cols):
    ind = rows*array_shape[1] + cols
    ind[ind < 0] = -1
    ind[ind >= array_shape[0]*array_shape[1]] = -1
    return ind


def ind2sub(array_shape, ind):
    ind[ind < 0] = -1
    ind[ind >= array_shape[0]*array_shape[1]] = -1
    rows = (ind.astype('int') / array_shape[1])
    cols = ind % array_shape[1]
    return rows, cols


import numpy as np

def distancexform(occgrid, goal, metric='cityblock', show=False):
    """
    Distance transform for path planning

    :param occgrid: Occupancy grid
    :type occgrid: NumPy ndarray
    :param goal: Goal position (x,y)
    :type goal: 2-element array-like
    :param metric: distance metric, defaults to 'cityblock'
    :type metric: str, optional
    :return: Distance transform matrix
    :rtype: NumPy ndarray

    Returns a matrix the same size as the occupancy grid ``occgrid`` where each
    cell contains the distance to the goal according to the chosen ``metric``.

        - Obstacle cells will be set to ``nan``.  
        - Unreachable cells, ie. free cells _inside obstacles_ will be set 
          to ``inf``. 

    The cells of the passed occupancy grid are:
        - zero, cell is free or driveable
        - one, cell is an obstacle, or not driveable
    """

    # build the matrix for performing distance transform:
    # - obstacles are nan
    # - other cells are inf
    # - goal is zero

    goal = base.getvector(goal, 2, dtype=np.int)
    occgrid = occgrid.astype(np.float32)
    occgrid[occgrid > 0] = np.nan  # assign nan to obstacle cells
    nans = np.isnan(occgrid)  # keep record of where the NaNs are

    occgrid[occgrid==0] = np.inf  # assign inf to other cells
    occgrid[goal[1], goal[0]] = 0  # assign zero to goal
    
    # create the appropriate distance matrix D
    if metric.lower() in ('manhattan', 'cityblock'):
        D = np.array([
                [ np.inf,   1,   np.inf],
                [      1,   0,        1],
                [ np.inf,   1,   np.inf]
            ])
    elif metric.lower() == 'euclidean':
        r2 = np.sqrt(2)
        D = np.array([
                [ r2,   1,   r2],
                [  1,   0,    1],
                [ r2,   1,   r2]
                ])

    # get ready to iterate
    count = 0
    ninf = np.inf  # number of infinities in the map

    while True:
        
        occgrid = dxstep(occgrid, D)
        occgrid[nans] = np.nan

        count += 1
        # if opt.animate
        #     cmap = [1 0 0; gray(count)];
        #     colormap(cmap)
        #     image(occgrid+1, 'CDataMapping', 'direct');
        #     set(gca, 'Ydir', 'normal');
        #     xlabel('x');
        #     ylabel('y');
        #     if opt.animate
        #         anim.add();
        #     else
        #         pause(opt.delay);

        ninfnow = sum(np.isinf(occgrid.flatten()))  # current number of Infs
        if ninfnow == ninf:
            # stop if the number of Infs left in the map had stopped reducing
            # it may never get to zero if there are unreachable cells in the map
            break

        ninf = ninfnow

    print(f"{count:d} iterations, {ninf:d} unreachable cells")
    return occgrid

def dxstep(G, D):

    # pad with inf
    H = np.pad(G, max(D.shape) // 2, 'constant', constant_values=np.inf)
    rows, columns = G.shape
    
    # inspired by https://landscapearchaeology.org/2018/numpy-loops/
    minimum = np.full(G.shape, np.inf)
    for y in range(3):
        for x in range(3):
            v = H[y : rows + y, x : columns + x] + D[y, x]
            # we use fmin() because it ignores NaNs, behaves like MATLAB min()
            minimum = np.fmin(minimum, v)

    return minimum


if __name__ == "__main__":


    # # make a simple map, as per the MOOCs
    # occgrid = np.zeros((10,10))
    # occgrid[3:6,2:7] = 1
    # occgrid[4,3:5] = 0  # hole in the obstacle
    # # occgrid[7:8,7] = 1  # extra bit

    # print(occgrid)
    # print()

    # goal=[5,7]
    # np.set_printoptions(precision=1)
    # dx = distancexform(occgrid, goal, metric='Euclidean')
    # print(dx)


    from roboticstoolbox import DXform
    from scipy.io import loadmat

    vars = loadmat("/Users/corkep/code/robotics-toolbox-python/data/house.mat", squeeze_me=True, struct_as_record=False)
    house = vars['house']
    place = vars['place']

    dx = DXform(house)
    print(dx)
    dx.goal = [1,2]
    dx.plan(place.kitchen)
    dx.plot()


#!/usr/bin/env python
"""
@author Jesse Haviland
"""

# import numpy as np
from roboticstoolbox.backends.PyPlot.RobotPlot import RobotPlot


class RobotPlot2(RobotPlot):

    def __init__(
            self, robot, ax, readonly, display=True,
            eeframe=True, name=True):

        super(RobotPlot2, self).__init__(
            robot, ax, readonly, display=display,
            jointaxes=False, shadow=False, eeframe=eeframe, name=name
        )



"""
Python Bug Planner
@Author: Kristian Gibson
TODO: Comments + Sphynx Docs Structured Text
TODO: Bug-fix, testing

Not ready for use yet.
"""
from numpy import disp
from scipy import integrate

from spatialmath.pose2d import SE2
from spatialmath import base
from spatialmath.base.animate import *
from scipy.ndimage import *
from matplotlib import cm
from roboticstoolbox.mobile.navigation import Navigation


class Bug2(Navigation):
    def __init__(self, occ_grid, **kwargs):

        """
        
                    %Bug2.Bug2 Construct a Bug2 navigation object 
            %
            % B = Bug2(MAP, OPTIONS) is a bug2 navigation object, and MAP is an occupancy grid,
            % a representation of a planar world as a matrix whose elements are 0 (free
            % space) or 1 (occupied).
            %
            % Options::
            % 'goal',G      Specify the goal point (1x2)
            % 'inflate',K   Inflate all obstacles by K cells.
            %
            % See also Navigation.Navigation.

                    %Navigation.Navigation Create a Navigation object
        %
        % N = Navigation(OCCGRID, OPTIONS) is a Navigation object that holds an
        % occupancy grid OCCGRID.  A number of options can be be passed.
        %
        % Options::
        % 'goal',G        Specify the goal point (2x1)
        % 'inflate',K     Inflate all obstacles by K cells.
        % 'private'       Use private random number stream.
        % 'reset'         Reset random number stream.
        % 'verbose'       Display debugging information
        % 'seed',S        Set the initial state of the random number stream.  S must
        %                 be a proper random number generator state such as saved in
        %                 the seed0 property of an earlier run.
        %
        % Notes::
        % - In the occupancy grid a value of zero means free space and non-zero means
        %   occupied (not driveable).
        % - Obstacle inflation is performed with a round structuring element (kcircle) 
        %   with radius given by the 'inflate' option.
        % - Inflation requires either MVTB or IPT installed.
        % - The 'private' option creates a private random number stream for the methods 
        %   rand, randn and randi.  If not given the global stream is used.
        %
        % See also randstream.

        """
        super().__init__(occ_grid=occ_grid, **kwargs)

        self._h = []
        self._j = 0
        self._step = 1
        self._m_line = None
        self._edge = None
        self._k = None


    @property
    def h(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        return self._h

    @property
    def j(self):
        return self._j

    @property
    def step(self):
        return self._step

    @property
    def m_line(self):
        return self._m_line

    @property
    def edge(self):
        return self._edge

    @property
    def k(self):
        return self._k

    def query(self, start=None, goal=None, animate=False, movie=None, current=False):
        """
        [summary]

        :param start: [description], defaults to None
        :type start: [type], optional
        :param goal: [description], defaults to None
        :type goal: [type], optional
        :param animate: [description], defaults to False
        :type animate: bool, optional
        :param movie: [description], defaults to None
        :type movie: [type], optional
        :param current: [description], defaults to False
        :type current: bool, optional
        :return: [description]
        :rtype: [type]

                    %Bug2.query  Find a path
            %
            % B.query(START, GOAL, OPTIONS) is the path (Nx2) from START (1x2) to GOAL
            % (1x2).  Row are the coordinates of successive points along the path.  If
            % either START or GOAL is [] the grid map is displayed and the user is
            % prompted to select a point by clicking on the plot.
            %
            % Options::
            %  'animate'   show a simulation of the robot moving along the path
            %  'movie',M   create a movie
            %  'current'   show the current position position as a black circle
            %
            % Notes::
            % - START and GOAL are given as X,Y coordinates in the grid map, not as
            %   MATLAB row and column coordinates.
            % - START and GOAL are tested to ensure they lie in free space.
            % - The Bug2 algorithm is completely reactive so there is no planning
            %   method.
            % - If the bug does a lot of back tracking it's hard to see the current
            %   position, use the 'current' option.
            % - For the movie option if M contains an extension a movie file with that
            %   extension is created.  Otherwise a folder will be created containing
            %   individual frames.
            %
            % See also Animate.
        """
        anim = None
        if movie is not None:
            anim = Animate(movie)
            animate = True

        # make sure start and goal are set and valid
        self.check_query(start, goal)

        # compute the m-line
        #  create homogeneous representation of the line
        #  line*[x y 1]' = 0
        self._m_line = hom_line(self._start[0], self._start[1],
                                   self._goal[0], self._goal[1])

        if animate:
            self.plot()
            self.plot_m_line()

        robot = self._start[:]
        self._step = 1
        path = [self._start]
        h = None

        # iterate using the next() method until we reach the goal
        while True:
            if animate:
                plt.plot(robot[0], robot[1], 'y.')
                plt.pause(0.1)
                if current:
                    h = self.plot(robot[0], robot[1])
                plt.draw()
                if movie is not None:
                    anim.plot(h)
                if current:
                    self.delete(h)

            # move to next point on path
            robot = self.next(robot)

            # have we been here before, ie. in a loop
            if any([all(robot == x) for x in path]):
                raise RuntimeError('trapped')

            # are we there yet?
            if robot is None:
                break
            else:
                path.append(robot)

        if movie is not None:
            anim.done()

        return np.r_[path]

    def plot_m_line(self, ls=None):
        if ls is None:
            ls = 'k--'

        x_min, x_max = plt.gca().get_xlim()
        y_min, y_max = plt.gca().get_ylim()
        if self._m_line[1] == 0:
            # handle the case that the line is vertical
            plt.plot([self._start[0], self._start[0]],
                     [y_min, y_max], 'k--')
        else:
            # regular line
            x = np.array([
                [x_min, 1],
                [x_max, 1]  ])
            y = -x @ np.r_[self._m_line[0], self._m_line[2]]
            y = y / self._m_line[1]
            plt.plot([x_min, x_max], y, ls, zorder=10)

    def next(self, position):

        l = None
        y = None

        if all(self._goal == position):
            return None  # we have arrived

        if self._step == 1:
            # Step 1.  Move along the M-line toward the goal
            self.message(f"{position}: moving along the M-line (step 1)")
            # motion on line toward goal
            d = self._goal - position
            if abs(d[0]) > abs(d[1]):
                # line slope less than 45 deg
                dx = 1 if d[0] >= 0 else -1  # np.sign(d[0])
                l = self._m_line
                y = -((position[0] + dx) * l[0] + l[2]) / l[1]
                dy = int(round(y - position[1]))
            else:
                # line slope greater than 45 deg
                dy = 1 if d[1] >= 0 else -1  # np.sign(d[1])
                l = self._m_line
                x = -((position[1] + dy) * l[1] + l[2]) / l[0]
                dx = int(round(x - position[0]))

            # detect if next step is an obstacle
            if self.is_occupied(position + np.r_[dx, dy]):
                self.message(f"  {position}: obstacle at {position + np.r_[dx, dy]}")
                self.h.append(position)
                self._step = 2  # transition to step 2
                self.message(f"  {position}: change to step 2")

                # get a list of all the points around the obstacle
                self._edge, _ = edgelist(self._occ_grid_nav == 0, position)
                self._k = 1
            else:
                n = position + np.array([dx, dy])

        if self._step == 2:
            # Step 2.  Move around the obstacle until we reach a point
            # on the M-line closer than when we started.

            self.message(f"{position}: moving around the obstacle (step 2)")
            if self._k <= len(self._edge):
                n = self._edge[self._k]  # next edge point
            else:
                # we are at the end of the list of edge points, we
                # are back where we started.  Step 2.c test.
                raise RuntimeError('robot is trapped')

            # are we on the M-line now ?
            if abs(np.inner(np.r_[position, 1], self._m_line)) <= 0.5:
                self.message(f"  {position}: crossed the M-line")

                # are we closer than when we encountered the obstacle?
                if base.norm(position - self._goal) < base.norm(self._h[-1] - self._goal):
                    # self._j += 1
                    self._step = 1  # transition to step 1
                    self.message(f"  {position}: change to step 1")
                    return n

            # no, keep going around
            self._k += 1

        return n
        
    def plan(self):
        Error('RTB:Bug2:badcall', 'This class has no plan method')

# Ported from Peter Corke's edgelist function found:
# https://github.com/petercorke/toolbox-common-matlab/blob/master/edgelist.m

#  these are directions of 8-neighbours in a clockwise direction
_dirs = np.array([
        [-1,  0],
        [-1,  1],
        [ 0,  1],
        [ 1,  1],
        [ 1,  0],
        [ 1, -1],
        [ 0, -1],
        [-1, -1],
    ])

def edgelist(im, p, direction=1):
    """
    Find edge of a region

    :param im: binary image
    :type im: ndarray(h,w,int)
    :param p: initial point
    :type p: array_like(2)
    :param direction: direction to traverse region, +1 clockwise [default], -1
        counter-clockwise
    :type direction: int, optional
    :raises ValueError: initial point is not on the edge
    :raises RuntimeError: not able to find path to the goal
    :return: edge list, direction vector list
    :rtype: tuple of lists

    ``edge, dirs = edgelist(im, seed)`` is the boundary/contour/edge of a region
    in the binary image ``im``.  ``seed=[X,Y]`` is the coordinate of a point on
    the edge of the region of interest, but belonging to the region.

    ``edge`` is a list of coordinates (2) of edge pixels of a region in theThe
    elements of the edgelist are NumPy ndarray(2).

    ``dirs`` is a list of integers representing the direction of the edge from
    the corresponding point in ``edge`` to the next point in ``edge``.  The
    integers in the range 0 to 7 represent directions: W SW S SE E NW N NW
    respectively.

    - Coordinates are given and returned assuming the matrix is an image, so the
      indices are always in the form (x,y) or (column,row).
    - ``im` is a binary image where 0 is assumed to be background, non-zero 
      is an object.
    - ``p`` must be a point on the edge of the region.
    - The initial point is always the first and last element of the returned edgelist.
    - 8-direction chain coding can give incorrect results when used with
      blobs founds using 4-way connectivty.

    :Reference:

    - METHODS TO ESTIMATE AREAS AND PERIMETERS OF BLOB-LIKE OBJECTS: A COMPARISON
      Luren Yang, Fritz Albregtsen, Tor Lgnnestad and Per Grgttum
      IAPR Workshop on Machine Vision Applications Dec. 13-15, 1994, Kawasaki

    """

    if direction > 0:
        neighbours = np.arange(start=0, stop=8, step=1)
    else:
        neighbours = np.arange(start=7, stop=-1, step=-1)

    try:
        pix0 = im[p[1], p[0]]  # color of pixel we start at
    except:
        raise ValueError('specified coordinate is not within image')

    p = base.getvector(p, 2, dtype=np.int)
    q = adjacent_point(im, p, pix0)

    # find an adjacent point outside the blob
    if q is None:
        raise ValueError('no neighbour outside the blob')

    d = None
    e = [p]  # initialize the edge list
    dir = [] # initialize the direction list
    p0 = None

    while True:

        # find which direction is Q
        dq = q - p
        for kq in range(0, 8):
            # get index of neighbour's direction in range [1,8]
            if np.all(dq == _dirs[kq]):
                break

        # now test for directions relative to Q
        for j in neighbours:
            # get index of neighbour's direction in range [1,8]
            k = (j + kq) % 8
            # if k > 7:
            #     k = k - 7
            

            # compute coordinate of the k'th neighbour
            nk = p + _dirs[k]

            try:
                if im[nk[1], nk[0]] == pix0:
                    # if this neighbour is in the blob it is the next edge pixel
                    p = nk
                    break
            except:
                raise ValueError("Something went wrong calculating edgelist")

            q = nk

        dir.append(k)
        # check if we are back where we started
        if p0 is None:
                p0 = p  # make a note of where we started
        else:
            if all(p == p0):
                break

        # keep going, add this point to the edgelist
        e.append(p)
    

    return e, dir

# Ported from Peter Corke's adjacent_point function found:
# https://github.com/petercorke/toolbox-common-matlab/blob/master/edgelist.m

def adjacent_point(im, seed, pix0):
    """Find adjacent point

    :param im: occupancy grid
    :type im: ndarray(m,n)
    :param seed: initial point
    :type seed: ndarray(2)
    :param pix0: value of occupancy grid at ``seed`` coordinate
    :type pix0: int
    :return: coordinate of a neighbour
    :rtype: ndarray(2) or None

    Is a neighbouring point of the coordinate ``seed`` that is not within the
    region containing the coordinate ``seed``, ie. it is a neighbour but
    outside.
    """
    p = None

    for d in _dirs:
        p = seed + d
        try:
            if im[p[1], p[0]] != pix0:
                return p
        except:
            pass 

    return None


# Implementation of Peter Corke's matlab homline function from:
# https://github.com/petercorke/spatialmath-matlab/blob/master/homline.m
def hom_line(x1, y1, x2, y2):
    line = np.cross(np.r_[x1, y1, 1], np.r_[x2, y2, 1])

    # normalize so that the result of x*l' is the pixel distance
    # from the line
    return line / np.linalg.norm(line[0:2])

# # Sourced from: https://stackoverflow.com/questions/28995146/matlab-ind2sub-equivalent-in-python/28995315#28995315
# def sub2ind(array_shape, rows, cols):
#     ind = rows*array_shape[1] + cols
#     ind[ind < 0] = -1
#     ind[ind >= array_shape[0]*array_shape[1]] = -1
#     return ind


# def ind2sub(array_shape, ind):
#     ind[ind < 0] = -1
#     ind[ind >= array_shape[0]*array_shape[1]] = -1
#     rows = (ind.astype('int') / array_shape[1])
#     cols = ind % array_shape[1]
#     return rows, cols

# def col_norm(x):
#     y = np.array([])
#     if x.ndim > 1:
#         x = np.column_stack(x)
#         for vector in x:
#             y = np.append(y, np.linalg.norm(vector))
#     else:
#         y = np.linalg.norm(x)
#     return y

if __name__ == "__main__":

    from roboticstoolbox import loadmat

    vars = loadmat("../data/map1.mat")
    map = vars['map']

    bug = Bug2(map)
    # bug.plan()
    path = bug.query([20, 10], [50, 35])


#!/usr/bin/env python
# Created by: Aditya Dua
# 5 October 2017

from math import pi
import numpy as np
from roboticstoolbox import DHRobot, RevoluteDH
from spatialmath import SE3


class Orion5(DHRobot):
    """
    Class that models a RAWR robotics Orion5 manipulator

    ``Orion5()`` is a class which models a RAWR Robotics Orion5 robot and
    describes its kinematic characteristics using standard DH
    conventions.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.DH.Orion5()
        >>> print(robot)

    Defined joint configurations are:

        - qz, zero angles, all folded up
        - qv, stretched out vertically
        - qh, arm horizontal, hand down

    .. note::

      - SI units of metres are used.
      - Robot has only 4 DoF.

    :references:
        - https://rawrobotics.com.au/orion5
        - https://drive.google.com/file/d/0B6_9_-ZgiRdTNkVqMEkxN2RSc2c/view

    .. codeauthor:: Aditya Dua
    .. codeauthor:: Peter Corke
    """

    def __init__(self, base=None):

        mm = 1e-3
        deg = pi / 180

        # details from

        h = 53.0 * mm
        r = 30.309 * mm
        l2 = 170.384 * mm
        l3 = 136.307 * mm
        l4 = 86.00 * mm
        c = 40.0 * mm

        # tool_offset = l4 + c

        # Turret, Shoulder, Elbow, Wrist, Claw
        links = [
                 RevoluteDH(d=h, a=0, alpha=90 * deg),  # Turret
                 RevoluteDH(
                    d=0, a=l2, alpha=0,         # Shoulder
                    qlim=[10 * deg, 122.5 * deg]),
                 RevoluteDH(
                    d=0, a=-l3, alpha=0,         # Elbow
                    qlim=[20 * deg, 340 * deg]),
                 RevoluteDH(
                    d=0, a=l4+c, alpha=0,        # Wrist
                    qlim=[45 * deg, 315 * deg])
            ]

        super().__init__(
            links,
            name="Orion 5",
            manufacturer="RAWR Robotics",
            base=SE3(r, 0, 0),
            tool=SE3.Ry(90, 'deg'))

        # zero angles, all folded up
        self.addconfiguration("qz", np.r_[0, 90, 180, 180] * deg)

        # stretched out vertically
        self.addconfiguration("qv", np.r_[0, 90, 180, 180] * deg)

        # arm horizontal, hand down
        self.addconfiguration("qh", np.r_[0, 0, 180, 90] * deg)


if __name__ == '__main__':   # pragma nocover

    orion = Orion5()
    print(orion)



#!/usr/bin/env python
"""
@author Micah Huth
"""


from vpython import radians, vector
from numpy import array
from spatialmath import SE3

"""
global variables that can be used to easily reference X, Y, and Z axis
directions.
"""
x_axis_vector = vector(1, 0, 0)
y_axis_vector = vector(0, 1, 0)
z_axis_vector = vector(0, 0, 1)


def get_pose_x_vec(se3_obj):  # pragma nocover
    """
    Convert SE3 details to VPython vector format

    :param se3_obj: SE3 pose and orientation object
    :type se3_obj: class:`spatialmath.pose3d.SE3`
    :return: VPython vector representation of the X orientation
    :rtype: class:`vpython.vector`
    """
    data = se3_obj.n
    return vector(data[0], data[1], data[2])


def get_pose_y_vec(se3_obj):  # pragma nocover
    """
    Convert SE3 details to VPython vector format

    :param se3_obj: SE3 pose and orientation object
    :type se3_obj: class:`spatialmath.pose3d.SE3`
    :return: VPython vector representation of the Y orientation
    :rtype: class:`vpython.vector`
    """
    data = se3_obj.o
    return vector(data[0], data[1], data[2])


def get_pose_z_vec(se3_obj):  # pragma nocover
    """
    Convert SE3 details to VPython vector format

    :param se3_obj: SE3 pose and orientation object
    :type se3_obj: class:`spatialmath.pose3d.SE3`
    :return: VPython vector representation of the Z orientation
    :rtype: class:`vpython.vector`
    """
    data = se3_obj.a
    return vector(data[0], data[1], data[2])


def get_pose_pos(se3_obj):  # pragma nocover
    """
    Convert SE3 details to VPython vector format

    :param se3_obj: SE3 pose and orientation object
    :type se3_obj: `spatialmath.pose3d.SE3`
    :return: VPython vector representation of the position
    :rtype: class:`vpython.vector`
    """
    data = se3_obj.t
    return vector(data[0], data[1], data[2])


def vpython_to_se3(graphic_object):  # pragma nocover
    """
    This function will take in a graphics object and output it's pose as an
    SE3 object

    :param graphic_object: A VPython graphic object
    :type graphic_object: class:`vpython.object`
    :return: SE3 representation of the pose
    :rtype: class:`spatialmath.pose3d.SE3`
    """
    # Get the x, y, z axes and position
    x_vec = graphic_object.axis
    y_vec = graphic_object.up
    z_vec = x_vec.cross(y_vec)
    pos = graphic_object.pos

    # Form a numpy array
    T = array([
        [x_vec.x, y_vec.x, z_vec.x, pos.x],
        [x_vec.y, y_vec.y, z_vec.y, pos.y],
        [x_vec.z, y_vec.z, z_vec.z, pos.z],
        [0,       0,       0,     1]
    ])

    return SE3(T)


def wrap_to_pi(angle_type, angle):  # pragma nocover
    """
    Wrap the given angle (deg or rad) to [-pi pi]

    :param angle_type: String of whether the angle is deg or rad
    :type angle_type: `str`
    :param angle: The angle to wrap
    :type angle: `float`
    :raises ValueError: Throws the error if the given string is not "deg"
        or "rad"
    :return: The wrapped angle
    :rtype: `float`
    """
    if angle_type == "deg":
        angle = angle % 360
        if angle > 180:
            angle -= 360

    elif angle_type == "rad":
        angle = angle % radians(360)
        if angle > radians(180):
            angle -= radians(360)

    else:
        raise ValueError('angle_type must be "deg" or "rad"')

    return angle


def close_localhost_session(canvas):  # pragma nocover
    """
    Terminate the local host session through JavaScript

    :param canvas: The scene to append the JS to the caption
    :type canvas:
        class:`roboticstoolbox.backends.VPython.graphics_canvas.GraphicsCanvas3D`,
        class:`roboticstoolbox.backends.VPython.graphics_canvas.GraphicsCanvas2D`
    """

    canvas.scene.append_to_caption('''
        <script type="text/javascript">
            close();
        </script>
        ''')



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ERobot import ERobot


class Frankie(ERobot):
    """
    Class that imports a Frankie URDF model

    ``Frankie()`` is a class which imports a Frankie robot definition
    from a URDF file.  The model describes its kinematic and graphical
    characteristics.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.URDF.Frankie()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration
    - qs, arm is stretched out in the x-direction
    - qn, arm is at a nominal non-singular configuration

    .. codeauthor:: Jesse Haviland
    .. sectionauthor:: Peter Corke
    """
    def __init__(self):

        elinks, name = super().urdf_to_ets_args(
            "franka_description/robots/frankie_arm_hand.urdf.xacro")

        super().__init__(
            elinks,
            name=name,
            manufacturer='Franka Emika',
            gripper_links=elinks[11]
        )

        self.qdlim = np.array([
            4.0, 4.0,
            2.1750, 2.1750, 2.1750, 2.1750, 2.6100, 2.6100, 2.6100,
            3.0, 3.0])

        self.addconfiguration("qz", np.array(
            [0, 0, 0, 0, 0, 0, 0, 0, 0]))

        self.addconfiguration("qr", np.array(
            [0, 0, 0, -0.3, 0, -2.2, 0, 2.0, np.pi/4]))


if __name__ == '__main__':   # pragma nocover

    robot = Frankie()
    print(robot)



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import numpy as np


def null(A, atol=1e-13, rtol=0):
    '''
    Compute an approximate basis for the nullspace of A.

    The algorithm used by this function is based on the singular value
    decomposition of A.

    :param A: A should be at most 2D. A 1-D array with length k will be
        treated as a 2D with shape (1, k)
    :type A: ndarray
    :param atol: The absolute tolerance for a zero singular value. Singular
        values smaller than `atol` are considered to be zero.
    :type atol: float
    :param rtol: The relative tolerance. Singular values less than rtol*smax
        are considered to be zero, where smax is the largest singular value.
    :type rtol: float

    :notes:
        - If both `atol` and `rtol` are positive, the combined tolerance is
          the maximum of the two; that is:
          tol = max(atol, rtol * smax)
        - Singular values smaller than `tol` are considered to be zero.

    Return value
    ------------
    :return ns: If A is an array with shape (m, k), then ns will be an array
        with shape (k, n), where n is the estimated dimension of the
        nullspace of A.  The columns of ns are a basis for the
        nullspace; each element in numpy.dot(A, ns) will be approximately
        zero.
    :rtype ns: ndarray

    '''

    A = np.atleast_2d(A)
    u, s, vh = np.linalg.svd(A)
    tol = max(atol, rtol * s[0])
    nnz = (s >= tol).sum()
    ns = vh[nnz:].conj().T
    return ns
